# -*- coding: utf-8 -*-
# File: box.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Implementation of BoundingBox class and related methods
"""

from dataclasses import dataclass
from math import ceil, floor
from typing import List, Optional, Sequence, no_type_check

import numpy as np
import numpy.typing as npt
from numpy import float32

from ..utils.detection_types import ImageType
from ..utils.file_utils import cocotools_available
from ..utils.logger import logger

if cocotools_available():
    import pycocotools.mask as coco_mask


# taken from https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/common.py


def coco_iou(box_a: npt.NDArray[float32], box_b: npt.NDArray[float32]) -> npt.NDArray[float32]:
    """
    Calculate iou for two arrays of bounding boxes in xyxy format

    :param box_a: Array of shape Nx4
    :param box_b: Array of shape Mx4

    :return: Array of shape NxM
    """

    def to_xywh(box: npt.NDArray[float32]) -> npt.NDArray[float32]:
        box = box.copy()
        box[:, 2] -= box[:, 0]
        box[:, 3] -= box[:, 1]
        return box

    ret = coco_mask.iou(to_xywh(box_a), to_xywh(box_b), np.zeros((len(box_b),), dtype=bool))
    # can accelerate even more, if using float32
    return ret.astype("float32")


# taken from https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/utils/np_box_ops.py


def area(boxes: npt.NDArray[float32]) -> npt.NDArray[float32]:
    """
    Computes area of boxes.

    :param boxes: numpy array with shape [N, 4] holding N boxes in xyxy format

    :return: a numpy array with shape [N*1] representing box areas
    """
    return np.array((boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1]), dtype=float32)


# taken from https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/utils/np_box_ops.py


def intersection(boxes1: npt.NDArray[float32], boxes2: npt.NDArray[float32]) -> npt.NDArray[float32]:
    """
    Compute pairwise intersection areas between boxes.

    :param boxes1: a numpy array with shape [N, 4] holding N boxes in xyxy format
    :param  boxes2: a numpy array with shape [M, 4] holding M boxes in xyxy format

    :return: a numpy array with shape [N*M] representing pairwise intersection area
    """
    [y_min1, x_min1, y_max1, x_max1] = np.split(boxes1, 4, axis=1)  # pylint: disable=W0632
    [y_min2, x_min2, y_max2, x_max2] = np.split(boxes2, 4, axis=1)  # pylint: disable=W0632

    all_pairs_min_ymax = np.minimum(y_max1, np.transpose(y_max2))
    all_pairs_max_ymin = np.maximum(y_min1, np.transpose(y_min2))
    intersect_heights = np.maximum(
        np.zeros(all_pairs_max_ymin.shape, dtype="f4"), all_pairs_min_ymax - all_pairs_max_ymin
    )
    all_pairs_min_xmax = np.minimum(x_max1, np.transpose(x_max2))
    all_pairs_max_xmin = np.maximum(x_min1, np.transpose(x_min2))
    intersect_widths = np.maximum(
        np.zeros(all_pairs_max_xmin.shape, dtype="f4"), all_pairs_min_xmax - all_pairs_max_xmin
    )
    return intersect_heights * intersect_widths


# taken from https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/utils/np_box_ops.py


def np_iou(boxes1: npt.NDArray[float32], boxes2: npt.NDArray[float32]) -> npt.NDArray[float32]:
    """
    Computes pairwise intersection-over-union between box collections.

    :param  boxes1: a numpy array with shape [N, 4] holding N boxes in xyxy format.
    :param  boxes2: a numpy array with shape [M, 4] holding M boxes in xyxy format.

    :return:  a numpy array with shape [N, M] representing pairwise iou scores.
    """
    intersect = intersection(boxes1, boxes2)
    area1 = area(boxes1)
    area2 = area(boxes2)
    union = np.expand_dims(area1, axis=1) + np.expand_dims(area2, axis=0) - intersect

    return intersect / union


def iou(boxes1: npt.NDArray[float32], boxes2: npt.NDArray[float32]) -> npt.NDArray[float32]:
    """
    Computes pairwise intersection-over-union between box collections. The method will be chosen based on what
    is installed:

    - If pycocotools is installed it will choose pycocotools.mask.iou which is a cython implementation and much faster
    - Otherwise it will use the numpy implementation as fallback

    :param boxes1:  a numpy array with shape [N, 4] holding N boxes in xyxy format.
    :param boxes2:  a numpy array with shape [N, 4] holding N boxes in xyxy format.

    :return: a numpy array with shape [N, M] representing pairwise iou scores.
    """

    if cocotools_available():
        return coco_iou(boxes1, boxes2)
    return np_iou(boxes1, boxes2)


class BoundingBoxError(BaseException):
    """Special exception only for `BoundingBox`"""


@dataclass
class BoundingBox:
    """
    Rectangular bounding box dataclass for object detection. Store coordinates and allows several
    representations. You can define an instance by passing the upper left point along with either height
    and width or along with the lower right point. Pass absolute_coords = 'True' if you work with image
    pixel coordinates. If you work with coordinates in the range between (0,1) then pass absolute_coords
    ='False'. A bounding box is a disposable object. Do not change the coordinates once the have been set but define
    a new box.

    `absolute_coords` indicates, whether given coordinates are in absolute or in relative terms

    `ulx`: upper left x

    `uly`: upper left y

    `lrx`: lower right x

    `lry`: lower right y

    `height`: height

    `width`: width
    """

    absolute_coords: bool
    ulx: float
    uly: float
    lrx: float = 0.0
    lry: float = 0.0
    height: float = 0.0
    width: float = 0.0

    def __post_init__(self) -> None:
        if self.width == 0.0:
            if self.lrx is None:
                raise BoundingBoxError("Bounding box not fully initialized")
            self.width = self.lrx - self.ulx
        if self.height == 0.0:
            if self.lry is None:
                raise BoundingBoxError("Bounding box not fully initialized")
            self.height = self.lry - self.uly

        if self.lrx == 0.0:
            if self.width is None:
                raise BoundingBoxError("Bounding box not fully initialized")
            self.lrx = self.ulx + self.width
        if self.lry == 0.0:
            if self.height is None:
                raise BoundingBoxError("Bounding box not fully initialized")
            self.lry = self.uly + self.height

        if not (self.ulx >= 0.0 and self.uly >= 0.0):
            raise BoundingBoxError("Bounding box ul must be >= (0.,0.)")
        if not (self.height > 0.0 and self.width > 0.0):
            raise BoundingBoxError(
                f"bounding box must have height and width >0. Check coords "
                f"ulx: {self.ulx}, uly: {self.uly}, lrx: {self.lrx}, "
                f"lry: {self.lry}."
            )
        if not self.absolute_coords:
            if not (self.ulx <= 1.0 and self.uly <= 1.0 and self.lrx <= 1.0 and self.lry <= 1.0):
                raise BoundingBoxError("coordinates must be between 0 and 1")

    @property
    def cx(self) -> float:
        """
        Bounding box center x coordinate
        """
        return self.ulx + 0.5 * self.width

    @property
    def cy(self) -> float:
        """
        Bounding box center y coordinate
        """
        return self.uly + 0.5 * self.height

    @property
    def center(self) -> List[float]:
        """
        Bounding box center [x,y]
        """
        return [self.cx, self.cy]

    @property
    def area(self) -> float:
        """
        Bounding box area
        """
        if self.absolute_coords:
            return self.width * self.height
        raise ValueError("Cannot calculate area, when bounding box coords are relative")

    def to_np_array(self, mode: str, scale_x: float = 1.0, scale_y: float = 1.0) -> npt.NDArray[np.float32]:
        """
        Returns the coordinates as numpy array.

        :param mode: Mode for coordinate arrangement:
                     "xyxy" for upper left/lower right point representation,
                     "xywh" for upper left and width/height representation or
                     "poly" for full eight coordinate polygon representation. x,y coordinates will be
                      returned in counter-clockwise order.

        :param scale_x: rescale the x coordinate. Defaults to 1
        :param scale_y: rescale the y coordinate. Defaults to 1
        :return: box coordinates
        """
        np_box_scale = np.array([scale_x, scale_y, scale_x, scale_y], dtype=np.float32)
        np_poly_scale = np.array(
            [scale_x, scale_y, scale_x, scale_y, scale_x, scale_y, scale_x, scale_y], dtype=np.float32
        )
        assert mode in ("xyxy", "xywh", "poly"), "Not a valid mode"
        if mode == "xyxy":
            return np.array([self.ulx, self.uly, self.lrx, self.lry], dtype=np.float32) * np_box_scale
        if mode == "xywh":
            return np.array([self.ulx, self.uly, self.width, self.height], dtype=np.float32) * np_box_scale
        return (
            np.array([self.ulx, self.uly, self.lrx, self.uly, self.lrx, self.lry, self.ulx, self.lry], dtype=np.float32)
            * np_poly_scale
        )

    def to_list(self, mode: str, scale_x: float = 1.0, scale_y: float = 1.0) -> List[float]:
        """
        Returns the coordinates as list

        :param mode:  Mode for coordinate arrangement:
                     "xyxy" for upper left/lower right point representation,
                     "xywh" for upper left and width/height representation or
                     "poly" for full eight coordinate polygon representation. x,y coordinates will be
                      returned in counter-clockwise order.

        :param scale_x: rescale the x coordinate. Defaults to 1
        :param scale_y: rescale the y coordinate. Defaults to 1
        :return: box coordinates
        """
        assert mode in ("xyxy", "xywh", "poly"), "Not a valid mode"
        if mode == "xyxy":
            return [
                self.ulx * scale_x,
                self.uly * scale_y,
                self.lrx * scale_x,
                self.lry * scale_y,
            ]
        if mode == "xywh":
            return [
                self.ulx * scale_x,
                self.uly * scale_y,
                self.width * scale_x,
                self.height * scale_y,
            ]
        return [
            self.ulx * scale_x,
            self.uly * scale_y,
            self.lrx * scale_x,
            self.uly * scale_y,
            self.lrx * scale_x,
            self.lry * scale_y,
            self.ulx * scale_x,
            self.lry * scale_y,
        ]

    def transform(
        self,
        image_width: float,
        image_height: float,
        absolute_coords: bool = False,
    ) -> "BoundingBox":
        """
        Transforms bounding box coordinates into absolute or relative coords. Internally, a new bounding box will be
        created. Changing coordinates requires width and height of the whole image.

        :param image_width: The horizontal image size
        :param image_height: The vertical image size
        :param absolute_coords: Whether to recalculate into absolute coordinates.

        :return: Either a list or np.array.
        """

        if absolute_coords != self.absolute_coords:  # only transforming in this case
            if self.absolute_coords:
                transformed_box = BoundingBox(
                    absolute_coords=not self.absolute_coords,
                    ulx=self.ulx / image_width,
                    uly=self.uly / image_height,
                    lrx=self.lrx / image_width,
                    lry=self.lry / image_height,
                )
            else:
                transformed_box = BoundingBox(
                    absolute_coords=not self.absolute_coords,
                    ulx=self.ulx * image_width,
                    uly=self.uly * image_height,
                    lrx=self.lrx * image_width,
                    lry=self.lry * image_height,
                )
            return transformed_box
        return self

    def __str__(self) -> str:
        return f"Bounding Box ulx: {self.ulx}, uly: {self.uly}, lrx: {self.lrx}, lry: {self.lry}"

    @staticmethod
    def remove_keys() -> List[str]:
        """
        A list of attributes to suspend from as_dict creation.
        """
        return ["height", "width"]

    @classmethod
    @no_type_check
    def from_dict(cls, **kwargs) -> "BoundingBox":
        """
        Create `BoundingBox` instance from dict

        :param kwargs: dict with  `BoundingBox` attributes
        :return: Initialized BoundingBox
        """
        return cls(**kwargs)


def intersection_box(
    box_1: BoundingBox, box_2: BoundingBox, width: Optional[float] = None, height: Optional[float] = None
) -> BoundingBox:
    """
    Returns the intersection bounding box of two boxes. Will raise a `ValueError` if the intersection is empty.
    If coords are absolute, it will floor the lower and ceil the upper coord to ensure the resulting box has same
    coordinates as the box induces from `crop_box_from_image`

    :param box_1: bounding box
    :param box_2: bounding box
    :param width: Total width of image. This optional parameter is needed if the value of `absolute_coords` of `box_1`
                  and `box_2` are not equal.
    :param height: Total height of image. This optional parameter is needed if the value of `absolute_coords` of `box_1`
                   and `box_2` are not equal.

    :return: bounding box. Will have same `absolute_coords` as `box_2`, if absolute_coords of `box_1` and `box_2` are
             not equal
    """

    if box_1.absolute_coords != box_2.absolute_coords:
        # will transform box_1
        assert (
            width is not None and height is not None
        ), "when absolute coords of boxes are not equal must pass width and height"
        box_1 = box_1.transform(width, height, box_2.absolute_coords)
    ulx = max(box_1.ulx, box_2.ulx)
    uly = max(box_1.uly, box_2.uly)
    lrx = min(box_1.lrx, box_2.lrx)
    lry = min(box_1.lry, box_2.lry)
    if box_2.absolute_coords:
        ulx, uly, lrx, lry = np.floor(ulx), np.floor(uly), np.ceil(lrx), np.ceil(lry)
    return BoundingBox(box_2.absolute_coords, ulx=ulx, uly=uly, lrx=lrx, lry=lry)


def crop_box_from_image(
    np_image: ImageType, crop_box: BoundingBox, width: Optional[float] = None, height: Optional[float] = None
) -> ImageType:
    """
    Crop a box (the crop_box) from a np_image. Will floor the left  and ceil the right coordinate point.

    :param np_image: Image to crop from.
    :param crop_box: Bounding box to crop.
    :param width: Total width of image. This optional parameter is needed if the value of absolute_coords of crop_box is
                  False

    :param height:Total width of image. This optional parameter is needed if the value of absolute_coords of crop_box is
                  False

    :return: A numpy array cropped according to the bounding box.
    """
    if not crop_box.absolute_coords:
        assert (
            width is not None and height is not None
        ), "when crop_box has absolute coords set to False must pass width and height"
        absolute_coord_box = crop_box.transform(width, height, absolute_coords=True)
    else:
        absolute_coord_box = crop_box

    assert isinstance(absolute_coord_box, BoundingBox)
    np_max_y, np_max_x = np_image.shape[0:2]
    return np_image[
        int(floor(absolute_coord_box.uly)) : min(int(ceil(absolute_coord_box.lry)), np_max_y),
        int(floor(absolute_coord_box.ulx)) : min(int(ceil(absolute_coord_box.lrx)), np_max_x),
    ]


def local_to_global_coords(local_box: BoundingBox, embedding_box: BoundingBox) -> BoundingBox:
    """
    Transform coords in terms of a cropped image into global coords. The local box coords are given in terms of the
    embedding box. The global coords will be determined by transforming the upper left point (which is (0,0) in
    local terms) into the upper left point given by the embedding box. This will shift the ul point of the
    local box to ul + embedding_box.ul

    :param local_box: bounding box with coords in terms of an embedding (e.g. local coordinates)
    :param embedding_box: bounding box of the embedding.
    :return: bounding box with local box transformed to absolute coords
    """

    assert local_box.absolute_coords and embedding_box.absolute_coords, (
        f"absolute coords "
        f"(={local_box.absolute_coords} for local_box and embedding_box (={embedding_box.absolute_coords}) must be "
        f"True"
    )
    assert embedding_box.ulx is not None and embedding_box.uly is not None
    assert (
        local_box.ulx is not None
        and local_box.uly is not None
        and local_box.lrx is not None
        and local_box.lry is not None
    )
    return BoundingBox(
        absolute_coords=True,
        ulx=embedding_box.ulx + local_box.ulx,
        uly=embedding_box.uly + local_box.uly,
        lrx=embedding_box.ulx + local_box.lrx,
        lry=embedding_box.uly + local_box.lry,
    )


def global_to_local_coords(global_box: BoundingBox, embedding_box: BoundingBox) -> BoundingBox:
    """
    Transforming global bounding box coords into the coordinate system given by the embedding box. The transformation
    requires that the global bounding box coordinates lie completely within the rectangle of the embedding box.
    The transformation results from a shift of all coordinates given by the shift of the upper left point of the
    embedding box into (0,0).

    :param global_box: The bounding box to be embedded
    :param embedding_box: The embedding box. Must cover the global box completely.
    :return: Bounding box of the embedded box in local coordinates.
    """

    assert global_box.absolute_coords and embedding_box.absolute_coords, (
        f"absolute coords "
        f"(={global_box.absolute_coords} for local_box and embedding_box (={embedding_box.absolute_coords}) must be "
        f"True"
    )

    return BoundingBox(
        absolute_coords=True,
        ulx=max(global_box.ulx - embedding_box.ulx, 0),
        uly=max(global_box.uly - embedding_box.uly, 0),
        lrx=min(global_box.lrx - embedding_box.ulx, embedding_box.width),
        lry=min(global_box.lry - embedding_box.uly, embedding_box.height),
    )


def merge_boxes(*boxes: BoundingBox) -> BoundingBox:
    """
    Generating the smallest box containing an arbitrary tuple/list of boxes. This function is only implemented for boxes
    with absolute coords = "True".

    :param boxes: An arbitrary tuple/list of bounding boxes `BoundingBox` all having absolute_coords="True".
    """
    absolute_coords = boxes[0].absolute_coords
    assert all(box.absolute_coords == absolute_coords for box in boxes), "all boxes must have same absolute_coords"

    ulx = min(box.ulx for box in boxes)
    uly = min(box.uly for box in boxes)
    lrx = max(box.lrx for box in boxes)
    lry = max(box.lry for box in boxes)

    return BoundingBox(absolute_coords=absolute_coords, ulx=ulx, uly=uly, lrx=lrx, lry=lry)


def rescale_coords(
    box: BoundingBox,
    current_total_width: float,
    current_total_height: float,
    scaled_total_width: float,
    scaled_total_height: float,
) -> BoundingBox:
    """
    Generating a bounding box with scaled coordinates. Will rescale x coordinate with factor

    * (current_total_width/scaled_total_width),

    resp. y coordinate with factor

    * (current_total_height/scaled_total_height),

    while not changing anything if `absolute_coords` is set to False.

    :param box: BoudingBox to rescale
    :param current_total_width: absolute coords of width of image
    :param current_total_height: absolute coords of height of image
    :param scaled_total_width:  absolute width of rescaled image
    :param scaled_total_height: absolute height of rescaled image
    :return: rescaled BoundingBox
    """

    if not box.absolute_coords:
        return box
    scale_width = scaled_total_width / current_total_width
    scale_height = scaled_total_height / current_total_height

    return BoundingBox(
        absolute_coords=True,
        ulx=box.ulx * scale_width,
        uly=box.uly * scale_height,
        lrx=box.lrx * scale_width,
        lry=box.lry * scale_height,
    )


def intersection_boxes(boxes_1: Sequence[BoundingBox], boxes_2: Sequence[BoundingBox]) -> Sequence[BoundingBox]:
    """
    The multiple version of 'intersection_box': Given two lists of m and n bounding boxes, it will calculate the
    pairwise intersection of both groups. There will be at most mxn intersection boxes.

    :param boxes_1: sequence of m BoundingBox
    :param boxes_2: sequence of n BoundingBox
    :return: list of at most mxn BoundingBox
    """
    if boxes_1[0].absolute_coords != boxes_1[0].absolute_coords:
        raise ValueError("absolute_coords of boxes_1 and boxes_2 mus be equal")
    absolute_coords = boxes_1[0].absolute_coords
    boxes1 = np.array([box.to_list(mode="xyxy") for box in boxes_1])
    boxes2 = np.array([box.to_list(mode="xyxy") for box in boxes_2])
    [x_min1, y_min1, x_max1, y_max1] = np.split(boxes1, 4, axis=1)  # pylint: disable=W0632
    [x_min2, y_min2, x_max2, y_max2] = np.split(boxes2, 4, axis=1)  # pylint: disable=W0632

    ulys = np.maximum(y_min1, np.transpose(y_min2))
    lrys = np.minimum(y_max1, np.transpose(y_max2))
    intersect_heights = np.maximum(np.zeros(ulys.shape, dtype="f4"), lrys - ulys).flatten()

    ulxs = np.maximum(x_min1, np.transpose(x_min2))
    lrxs = np.minimum(x_max1, np.transpose(x_max2))
    intersect_widths = np.maximum(np.zeros(ulxs.shape, dtype="f4"), lrxs - ulxs).flatten()
    np_boxes_output = np.swapaxes(
        [ulxs.flatten(), ulys.flatten(), intersect_widths.flatten(), intersect_heights.flatten()], 1, 0
    )
    boxes_output = []
    for idx in range(np_boxes_output.shape[0]):
        try:
            boxes_output.append(
                BoundingBox(
                    ulx=np_boxes_output[idx][0],
                    uly=np_boxes_output[idx][1],
                    width=np_boxes_output[idx][2],
                    height=np_boxes_output[idx][3],
                    absolute_coords=absolute_coords,
                )
            )
        except BoundingBoxError:
            log_dict = {
                "ulx": np_boxes_output[idx][0],
                "uly": np_boxes_output[idx][1],
                "width": np_boxes_output[idx][2],
                "height": np_boxes_output[idx][3],
            }

            logger.warning("intersection_boxes error %s", "", log_dict)

    return boxes_output



# -*- coding: utf-8 -*-
# File: box_ops.py

# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")

"""
Taken from

<https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/utils/box_ops.py>

and

<https://github.com/tensorflow/models/blob/master/object_detection/core/box_list_ops.py>
"""

# pylint: disable=import-error
import tensorflow as tf
from tensorpack.tfutils.scope_utils import under_name_scope

# pylint: enable=import-error


@under_name_scope()
def area(boxes):
    """
    :param boxes: nx4 floatbox

    :return: n
    """

    x_min, y_min, x_max, y_max = tf.split(boxes, 4, axis=1)
    return tf.squeeze((y_max - y_min) * (x_max - x_min), [1])


@under_name_scope()
def pairwise_intersection(boxlist1, boxlist2):
    """Compute pairwise intersection areas between boxes.

    :param boxlist1: Nx4 floatbox
    :param boxlist2: Mx4

    :return: A tensor with shape [N, M] representing pairwise intersections
    """

    x_min1, y_min1, x_max1, y_max1 = tf.split(boxlist1, 4, axis=1)
    x_min2, y_min2, x_max2, y_max2 = tf.split(boxlist2, 4, axis=1)
    all_pairs_min_ymax = tf.minimum(y_max1, tf.transpose(y_max2))
    all_pairs_max_ymin = tf.maximum(y_min1, tf.transpose(y_min2))
    intersect_heights = tf.maximum(0.0, all_pairs_min_ymax - all_pairs_max_ymin)
    all_pairs_min_xmax = tf.minimum(x_max1, tf.transpose(x_max2))
    all_pairs_max_xmin = tf.maximum(x_min1, tf.transpose(x_min2))
    intersect_widths = tf.maximum(0.0, all_pairs_min_xmax - all_pairs_max_xmin)
    return intersect_heights * intersect_widths


@under_name_scope()
def pairwise_iou(boxlist1, boxlist2):
    """Computes pairwise intersection-over-union between box collections.

    :param boxlist1: Nx4 floatbox
    :param boxlist2: Mx4

    :return: A tensor with shape [N, M] representing pairwise iou scores.
    """

    intersections = pairwise_intersection(boxlist1, boxlist2)
    areas1 = area(boxlist1)
    areas2 = area(boxlist2)
    unions = tf.expand_dims(areas1, 1) + tf.expand_dims(areas2, 0) - intersections
    return tf.where(tf.equal(intersections, 0.0), tf.zeros_like(intersections), tf.truediv(intersections, unions))



# -*- coding: utf-8 -*-
# File: common.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for common pipeline components
"""
from copy import deepcopy
from typing import List, Literal, Optional, Sequence, Union

import numpy as np

from ..dataflow import DataFlow, MapData
from ..datapoint.image import Image
from ..datapoint.view import Page
from ..mapper.maputils import MappingContextManager
from ..mapper.match import match_anns_by_intersection
from ..utils.detection_types import JsonDict
from ..utils.file_utils import detectron2_available, pytorch_available, tf_available
from ..utils.settings import LayoutType, ObjectTypes, Relationships, TypeOrStr, get_type
from .base import PipelineComponent
from .registry import pipeline_component_registry

if tf_available():
    from ..mapper.tpstruct import tf_nms_image_annotations as nms_image_annotations

elif pytorch_available() and detectron2_available():
    from ..mapper.d2struct import pt_nms_image_annotations as nms_image_annotations


@pipeline_component_registry.register("ImageCroppingService")
class ImageCroppingService(PipelineComponent):
    """
    Crop sub images given by bounding boxes of some annotations. This service is not necessary for
    `ImageLayoutService` and is more intended for saved files where sub images are
    generally not stored.
    """

    def __init__(self, category_names: Union[TypeOrStr, Sequence[TypeOrStr]]):
        """
        :param category_names: A single name or a list of category names to crop
        """

        if isinstance(category_names, str):
            category_names = [category_names]
        self.category_names = [get_type(category_name) for category_name in category_names]
        super().__init__("image_crop")

    def serve(self, dp: Image) -> None:
        for ann in dp.get_annotation(category_names=self.category_names):
            dp.image_ann_to_image(ann.annotation_id, crop_image=True)

    def clone(self) -> "PipelineComponent":
        return self.__class__(self.category_names)

    def get_meta_annotation(self) -> JsonDict:
        return dict([("image_annotations", []), ("sub_categories", {}), ("relationships", {}), ("summaries", [])])


@pipeline_component_registry.register("MatchingService")
class MatchingService(PipelineComponent):
    """
    Objects of two object classes can be assigned to one another by determining their pairwise average. If this is above
    a limit, a relation is created between them.
    The parent object class (based on its category) and the child object class are defined for the service. A child
    relation is created in the parent class if the conditions are met.

    Either `iou` (intersection-over-union) or `ioa` (intersection-over-area) can be selected as the matching rule.

            # the following will assign word annotations to text and title annotation, provided that their ioa-threshold
            # is above 0.7. words below that threshold will not be assigned.

            match = MatchingService(parent_categories=["TEXT","TITLE"],child_categories="WORD",matching_rule="ioa",
                                    threshold=0.7)

            # Assigning means that text and title annotation will receive a relationship called "CHILD" which is a list
              of annotation ids of mapped words.
    """

    def __init__(
        self,
        parent_categories: Union[TypeOrStr, List[TypeOrStr]],
        child_categories: Union[TypeOrStr, List[TypeOrStr]],
        matching_rule: Literal["iou", "ioa"],
        threshold: float,
        use_weighted_intersections: bool = False,
        max_parent_only: bool = False,
    ) -> None:
        """
        :param parent_categories: list of categories to be used a for parent class. Will generate a child-relationship
        :param child_categories: list of categories to be used for a child class.
        :param matching_rule: "iou" or "ioa"
        :param threshold: iou/ioa threshold. Value between [0,1]
        :param use_weighted_intersections: This is currently only implemented for matching_rule 'ioa'. Instead of using
                                           the ioa_matrix it will use mat weighted ioa in order to take into account
                                           that intersections with more cells will likely decrease the ioa value. By
                                           multiplying the ioa with the number of all intersection for each child this
                                           value calibrate the ioa.
        :param max_parent_only: Will assign to each child at most one parent with maximum ioa
        """
        self.parent_categories = parent_categories
        self.child_categories = child_categories
        assert matching_rule in ["iou", "ioa"], "segment rule must be either iou or ioa"
        self.matching_rule = matching_rule
        self.threshold = threshold
        self.use_weighted_intersections = use_weighted_intersections
        self.max_parent_only = max_parent_only
        super().__init__("matching")

    def serve(self, dp: Image) -> None:
        """
        - generates pairwise match-score by intersection
        - generates child relationship at parent level

        :param dp: datapoint image
        """
        child_index, parent_index, child_anns, parent_anns = match_anns_by_intersection(
            dp,
            parent_ann_category_names=self.parent_categories,
            child_ann_category_names=self.child_categories,
            matching_rule=self.matching_rule,
            threshold=self.threshold,
            use_weighted_intersections=self.use_weighted_intersections,
            max_parent_only=self.max_parent_only,
        )

        with MappingContextManager(dp_name=dp.file_name):
            matched_child_anns = np.take(child_anns, child_index)  # type: ignore
            matched_parent_anns = np.take(parent_anns, parent_index)  # type: ignore
            for idx, parent in enumerate(matched_parent_anns):
                parent.dump_relationship(Relationships.child, matched_child_anns[idx].annotation_id)

    def clone(self) -> PipelineComponent:
        return self.__class__(self.parent_categories, self.child_categories, self.matching_rule, self.threshold)

    def get_meta_annotation(self) -> JsonDict:

        return dict(
            [
                ("image_annotations", []),
                ("sub_categories", {}),
                ("relationships", {parent: {Relationships.child} for parent in self.parent_categories}),
                ("summaries", []),
            ]
        )


@pipeline_component_registry.register("PageParsingService")
class PageParsingService:
    """
    A "pseudo" pipeline component that can be added to a pipeline to convert `Image`s into `Page` formats. It allows a
    custom parsing depending on customizing options of other pipeline components.
    """

    def __init__(
        self,
        text_container: TypeOrStr,
        top_level_text_block_names: Union[TypeOrStr, Sequence[TypeOrStr]],
        text_block_names: Optional[Union[TypeOrStr, Sequence[TypeOrStr]]] = None,
    ):
        """
        :param text_container: name of an image annotation that has a CHARS sub category. These annotations will be
                               ordered within all text blocks.
        :param top_level_text_block_names: name of image annotation that have a relation with text containers (or which
                                           might be text containers themselves).
        :param text_block_names: name of image annotation that have a relation with text containers (or which might be
                                 text containers themselves). This is only necessary, when residual text_container (e.g.
                                 words that have not been assigned to any text block) should be displayed in `page.text`
        """
        if isinstance(top_level_text_block_names, (str, ObjectTypes)):
            top_level_text_block_names = [top_level_text_block_names]
        if isinstance(text_block_names, (str, ObjectTypes)):
            text_block_names = [text_block_names]
        if text_block_names is not None:
            text_block_names = [get_type(text_block) for text_block in text_block_names]

        self._text_container = get_type(text_container)
        self._top_level_text_block_names = [get_type(text_block) for text_block in top_level_text_block_names]
        self._text_block_names = text_block_names
        self._init_sanity_checks()

    def pass_datapoint(self, dp: Image) -> Page:
        """
        converts Image to Page
        :param dp: Image
        :return: Page
        """
        return Page.from_image(
            dp, self._text_container, self._top_level_text_block_names, self._text_block_names  # type: ignore
        )

    def predict_dataflow(self, df: DataFlow) -> DataFlow:
        """
        Mapping a datapoint via `pass_datapoint` within a dataflow pipeline

        :param df: An input dataflow
        :return: A output dataflow
        """
        return MapData(df, self.pass_datapoint)

    def _init_sanity_checks(self) -> None:
        assert self._text_container in [
            LayoutType.word,
            LayoutType.line,
        ], f"text_container must be either {LayoutType.word} or {LayoutType.line}"


@pipeline_component_registry.register("AnnotationNmsService")
class AnnotationNmsService(PipelineComponent):
    """
    A service to pass `ImageAnnotation` to a non-maximum suppression (NMS) process for given pairs of categories.
    `ImageAnnotation`s are subjected to NMS process in groups: If `nms_pairs=[[LayoutType.text, LayoutType.table],
    [LayoutType.title, LayoutType.table]]` all `ImageAnnotation` subject to these categories are being selected and
     identified as one category. After NMS the discarded image annotation will be deactivated.
    """

    def __init__(self, nms_pairs: Sequence[Sequence[TypeOrStr]], thresholds: Union[float, List[float]]):
        """
        :param nms_pairs: Groups of categories, either as string or by `ObjectType`.
        :param thresholds: Suppression threshold
        """
        self.nms_pairs = [[get_type(val) for val in pair] for pair in nms_pairs]
        if isinstance(thresholds, float):
            self.threshold = [thresholds for _ in self.nms_pairs]
        else:
            assert len(self.nms_pairs) == len(thresholds), "Sequences of nms_pairs and thresholds must have same length"
            self.threshold = thresholds
        super().__init__("nms")

    def serve(self, dp: Image) -> None:
        for pair, threshold in zip(self.nms_pairs, self.threshold):
            anns = dp.get_annotation(category_names=pair)
            ann_ids_to_keep = nms_image_annotations(anns, threshold, dp.image_id)
            for ann in anns:
                if ann.annotation_id not in ann_ids_to_keep:
                    self.dp_manager.deactivate_annotation(ann.annotation_id)

    def clone(self) -> "PipelineComponent":
        return self.__class__(deepcopy(self.nms_pairs), self.threshold)

    def get_meta_annotation(self) -> JsonDict:
        return dict([("image_annotations", []), ("sub_categories", {}), ("relationships", {}), ("summaries", [])])



# -*- coding: utf-8 -*-
# File: __init__.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Init module for train package
"""

from ..utils.file_utils import (
    detectron2_available,
    pytorch_available,
    tensorpack_available,
    tf_available,
    transformers_available,
)

if tf_available() and tensorpack_available():
    from .tp_frcnn_train import train_faster_rcnn

if pytorch_available() and detectron2_available():
    from .d2_frcnn_train import train_d2_faster_rcnn

if pytorch_available() and transformers_available():
    from .hf_detr_train import train_hf_detr
    from .hf_layoutlm_train import train_hf_layoutlm



# -*- coding: utf-8 -*-
# File: test_fastlang.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module extern.fastlang
"""
from typing import Tuple
from unittest.mock import MagicMock, patch

import numpy as np
import numpy.typing as npt
from numpy import float32
from pytest import mark

from deepdoctection.extern.fastlang import FasttextLangDetector
from deepdoctection.extern.model import ModelCatalog


def get_mock_lang_detect_result(text_string: str) -> Tuple[Tuple[str], npt.NDArray[float32]]:  # pylint: disable = W0613
    """return raw result from fasttext language detection model"""
    return (("__label__it",), np.array([0.99414486]))


class TestFasttextLangDetector:
    """
    Test FasttextLangDetector
    """

    @staticmethod
    @mark.requires_tf_or_pt
    @patch("deepdoctection.extern.fastlang.load_model", MagicMock(return_value=MagicMock()))
    def test_fasttext_lang_detector_predicts_language() -> None:
        """
        Detector calls model.predict(text_string) and processes returned results correctly
        """

        # Arrange
        path_weights = "/path/to/dir"
        profile = ModelCatalog.get_profile("fasttext/lid.176.bin")

        assert profile.categories
        fasttest_predictor = FasttextLangDetector(path_weights, profile.categories)
        fasttest_predictor.model.predict = MagicMock(side_effect=get_mock_lang_detect_result)

        # Act
        result = fasttest_predictor.predict("Un leggero dialetto italiano")

        # Assert
        assert result.text == "ita"
        assert result.score == 0.99414486



# -*- coding: utf-8 -*-
# File: laylmstruct.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for mapping annotations from image to layout lm input structure. Heavily inspired by the notebooks
<https://github.com/NielsRogge/Transformers-Tutorials>
"""

from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Literal, NewType, Optional, Sequence, Union

import numpy as np
import numpy.typing as npt
from cv2 import INTER_LINEAR

from ..datapoint.annotation import ContainerAnnotation
from ..datapoint.convert import box_to_point4, point4_to_box
from ..datapoint.image import Image
from ..utils.detection_types import JsonDict
from ..utils.file_utils import pytorch_available, transformers_available
from ..utils.settings import DatasetType, LayoutType, PageType, Relationships, WordType
from ..utils.transform import ResizeTransform, normalize_image
from .maputils import curry

if pytorch_available():
    import torch

if transformers_available():
    from transformers import (  # pylint: disable=W0611
        BatchEncoding,
        PreTrainedTokenizerFast,
        RobertaTokenizerFast,
        XLMRobertaTokenizerFast,
    )

__all__ = [
    "image_to_raw_layoutlm_features",
    "raw_features_to_layoutlm_features",
    "LayoutLMDataCollator",
    "image_to_layoutlm_features",
    "DataCollator",
    "LayoutLMFeatures",
]

RawLayoutLMFeatures = NewType("RawLayoutLMFeatures", JsonDict)
LayoutLMFeatures = NewType("LayoutLMFeatures", JsonDict)
InputDataClass = NewType("InputDataClass", JsonDict)

"""
<https://github.com/huggingface/transformers/src/transformers/data/data_collator.py>
A DataCollator is a function that takes a list of samples from a Dataset and collate them into a batch, as a dictionary
of PyTorch/TensorFlow tensors or NumPy arrays.
"""

DataCollator = NewType("DataCollator", Callable[[List[InputDataClass]], Dict[str, Any]])  # type: ignore

_CLS_BOX = [0.0, 0.0, 1000.0, 1000.0]
_SEP_BOX = [1000.0, 1000.0, 1000.0, 1000.0]


@curry
def image_to_raw_layoutlm_features(
    dp: Image,
    dataset_type: Optional[Literal["sequence_classification", "token_classification"]] = None,
    input_width: int = 1000,
    input_height: int = 1000,
    image_width: int = 1000,
    image_height: int = 1000,
    color_mode: Literal["BGR", "RGB"] = "BGR",
    pixel_mean: Optional[npt.NDArray[np.float32]] = None,
    pixel_std: Optional[npt.NDArray[np.float32]] = None,
    use_token_tag: bool = True,
    segment_positions: Optional[Union[LayoutType, Sequence[LayoutType]]] = None,
) -> Optional[RawLayoutLMFeatures]:
    """
    Mapping a datapoint into an intermediate format for layoutlm. Features will be provided into a dict and this mapping
    can be used for sequence or token classification as well as for inference. To generate input features for the model
    please `use raw_features_to_layoutlm_features`.


    :param dp: Image
    :param dataset_type: Either SEQUENCE_CLASSIFICATION or TOKEN_CLASSIFICATION. When using a built-in dataset use
    :param input_width: max width of box coordinates. Under the hood, it will transform the image and all box
                        coordinates accordingly.
    :param input_height: target height of box coordinates. Under the hood, it will transform the image and all box
                        coordinates accordingly.
    :param image_width: Some models (e.g. `Layoutlmv2`) assume box coordinates to be normalized to input_width, whereas
                        the image has to be resized to a different width. This input will only resize the `image` width.
    :param image_height: Some models (e.g. `Layoutlmv2`) assume box coordinates to be normalized to input_height,
                         whereas the image has to be resized to a different height. This input will only resize the
                         `image` height.
    :param color_mode: Either "BGR" or "RGB". Note, that LayoutLMv2 uses "BGR" because of Detectron2 backbone, whereas
                       LayoutLMv3 uses "RGB".
    :param pixel_mean: (3,) array for "BGR" resp. "RGB" mean
    :param pixel_std: (3,) array for "BGR" resp. "RGB" std
    :param use_token_tag: Will only be used for dataset_type="token_classification". If use_token_tag=True, will use
                          labels from sub category `WordType.token_tag` (with `B,I,O` suffix), otherwise
                          `WordType.token_class`.
    :param segment_positions: Using bounding boxes of segment instead of words improves model accuracy significantly.
                              Choose a single or a sequence of layout segments to use their bounding boxes. Note, that
                              the layout segments need to have a child-relationship with words. If a word does not
                              appear as child, it will use the word bounding box.
    :return: dictionary with the following arguments:
            'image_id', 'width', 'height', 'ann_ids', 'words', 'bbox' and 'dataset_type'.
    """

    raw_features: RawLayoutLMFeatures = RawLayoutLMFeatures({})
    all_ann_ids = []
    all_words = []
    all_boxes = []
    all_labels: List[int] = []

    anns = dp.get_annotation_iter(category_names=LayoutType.word)

    word_id_to_segment_box = {}
    if segment_positions:
        if isinstance(segment_positions, LayoutType):
            segment_positions = [segment_positions]
        segment_anns = dp.get_annotation(category_names=segment_positions)
        for segm_ann in segment_anns:
            if segm_ann.image is not None:
                bounding_box = segm_ann.image.get_embedding(dp.image_id)
            else:
                bounding_box = segm_ann.bounding_box
            if not bounding_box.absolute_coords:
                bounding_box = bounding_box.transform(dp.width, dp.height, absolute_coords=True)
            word_id_to_segment_box.update(
                {word_ann: bounding_box for word_ann in segm_ann.get_relationship(Relationships.child)}
            )

    for ann in anns:
        all_ann_ids.append(ann.annotation_id)
        char_cat = ann.get_sub_category(WordType.characters)
        if not isinstance(char_cat, ContainerAnnotation):
            raise TypeError(f"char_cat must be of type ContainerAnnotation but is of type {type(char_cat)}")
        word = char_cat.value
        if not isinstance(word, str):
            raise ValueError(f"word must be of type str but is of type {type(word)}")
        all_words.append(word)

        if ann.image is not None:
            box = ann.image.get_embedding(dp.image_id)
        else:
            box = ann.bounding_box
        assert box is not None, box
        if not box.absolute_coords:
            box = box.transform(dp.width, dp.height, absolute_coords=True)
        all_boxes.append(word_id_to_segment_box.get(ann.annotation_id, box).to_list(mode="xyxy"))

        if (
            WordType.token_tag in ann.sub_categories
            and WordType.token_class in ann.sub_categories
            and dataset_type == DatasetType.token_classification
        ):
            if use_token_tag:
                all_labels.append(int(ann.get_sub_category(WordType.token_tag).category_id) - 1)
            else:
                all_labels.append(int(ann.get_sub_category(WordType.token_class).category_id) - 1)

    if dp.summary is not None and dataset_type == DatasetType.sequence_classification:
        all_labels.append(int(dp.summary.get_sub_category(PageType.document_type).category_id) - 1)

    boxes = np.asarray(all_boxes, dtype="float32")
    if boxes.ndim == 1:
        return None

    boxes = box_to_point4(boxes)

    resizer = ResizeTransform(dp.height, dp.width, input_height, input_width, INTER_LINEAR)

    if dp.image is not None:
        if image_width != input_width or image_height != input_height:
            image_only_resizer = ResizeTransform(dp.height, dp.width, image_height, image_width, INTER_LINEAR)
            image = image_only_resizer.apply_image(dp.image)
        else:
            image = resizer.apply_image(dp.image)
        image_key = "image"
        if color_mode == "RGB":
            image = image[..., ::-1]
            image_key = "pixel_values"
        if pixel_mean is not None and pixel_std is not None:
            image = normalize_image(image, pixel_mean, pixel_std)
        raw_features[image_key] = image  # pylint: disable=E1137  #3162

    boxes = resizer.apply_coords(boxes)
    boxes = point4_to_box(boxes)

    # input box coordinates must be of type long. We floor the ul and ceil the lr coords
    boxes = np.concatenate((np.floor(boxes)[:, :2], np.ceil(boxes)[:, 2:]), axis=1).tolist()

    # pylint: disable=E1137  #3162
    raw_features["image_id"] = dp.image_id
    raw_features["width"] = input_width
    raw_features["height"] = input_height
    raw_features["ann_ids"] = all_ann_ids
    raw_features["words"] = all_words
    raw_features["bbox"] = boxes
    raw_features["dataset_type"] = dataset_type

    if all_labels:
        raw_features["labels"] = all_labels
    # pylint: enable=E1137
    return raw_features


def features_to_pt_tensors(features: LayoutLMFeatures) -> LayoutLMFeatures:
    """
    Converting list of floats to pytorch tensors
    :param features: LayoutLMFeatures
    :return: LayoutLMFeatures
    """

    _image_key = "pixel_values" if "pixel_values" in features else "image"
    features["bbox"] = torch.tensor(features["bbox"], dtype=torch.long)
    if "labels" in features:
        features["labels"] = torch.tensor(features["labels"], dtype=torch.long)
    if _image_key in features:
        features[_image_key] = torch.tensor(
            [image.astype("float32").transpose(2, 0, 1) for image in features[_image_key]], dtype=torch.float32
        )
        # features["images"] = [
        #    torch.as_tensor(image.astype("float32").transpose(2, 0, 1)) for image in features["images"]
        # ]
    return features


def _tokenize_with_sliding_window(
    raw_features: List[RawLayoutLMFeatures],
    tokenizer: "PreTrainedTokenizerFast",
    sliding_window_stride: int,
    return_tensors: Optional[Literal["pt"]] = None,
) -> Union[JsonDict, "BatchEncoding"]:
    """
    Runs a tokenizer: If there are no overflowing tokens, the tokenizer output will be returned as it is.
    If there are overflowing tokens, sliding windows have to be built. As it is easier to prepare the sliding windows
    from raw tokenized outputs we run the tokenizer a second time without truncating and build the sliding windows from
    this second output.
    """
    # first try: we require return_overflowing_tokens=True. If the number of raw features is equal to
    # overflow_to_sample_mapping then there is nothing more to do because the sample has less than max_length
    # tokens
    tokenized_inputs = tokenizer(
        [dp["words"] for dp in raw_features],
        padding="max_length",
        truncation=True,
        return_token_type_ids=True,
        return_overflowing_tokens=True,
        is_split_into_words=True,
        return_tensors=return_tensors,
    )
    if len(raw_features) == len(tokenized_inputs["overflow_to_sample_mapping"]):
        return tokenized_inputs

    # now we tokenize with neither truncation nor padding and build sliding windows.
    tokenized_inputs = tokenizer(
        [dp["words"] for dp in raw_features],
        padding="do_not_pad",
        truncation=False,
        return_token_type_ids=True,
        return_overflowing_tokens=False,
        is_split_into_words=True,
        return_tensors=None,
    )
    max_length = tokenizer.max_len_single_sentence  # here 510 (512 minus [CLS] and [SEP]
    sliding_windows_remainder = [
        divmod(len(sample.ids) - max_length, sliding_window_stride) for sample in tokenized_inputs.encodings
    ]  # list of (multiplier, remainder) per sample
    all_input_ids = []
    all_token_type_ids = []
    all_attention_mask = []
    all_word_ids = []
    all_tokens = []
    overflow_to_sample_mapping = []
    for idx, outputs in enumerate(
        zip(
            tokenized_inputs.encodings,
            tokenized_inputs.data["input_ids"],
            tokenized_inputs.data["token_type_ids"],
            tokenized_inputs.data["attention_mask"],
            sliding_windows_remainder,
        )
    ):
        encodings = outputs[0]
        input_ids_orig = outputs[1]
        token_type_ids_orig = outputs[2]
        attention_mask_orig = outputs[3]
        divmod_result = outputs[4]
        tokens_orig = encodings.tokens
        word_ids_orig = encodings.word_ids
        multiplier, _ = divmod_result
        total_length = len(tokens_orig)
        # suppose the sample has total_length= 525 tokens:
        #  [[CLS],1,...,523,[SEP]]. With sliding_window_stride = 8 we build windows as follows:
        # [[CLS],1,..,510,[SEP]], [[CLS],8,..,517,[SEP]], [[CLS],16,..,523,[SEP],[PAD],[PAD],[PAD]]
        # Here, we have a multiplier = 1,
        for k in range(multiplier + 2):
            start = max(k * sliding_window_stride, 1)
            end = min(max_length + start, total_length)
            pad_last = max(max_length + start - end, 0)
            overflow_to_sample_mapping.append(idx)
            if not pad_last:
                tokens = tokens_orig[start:end]
                tokens.insert(0, tokenizer.cls_token)
                tokens.append(tokenizer.sep_token)
                all_tokens.append(tokens)
                input_ids = input_ids_orig[start:end]
                input_ids.insert(0, 101)
                input_ids.append(102)
                all_input_ids.append(input_ids)
                token_type_ids = token_type_ids_orig[start:end]
                token_type_ids.insert(0, 0)
                token_type_ids.append(0)
                all_token_type_ids.append(token_type_ids)
                attention_mask = attention_mask_orig[start:end]
                attention_mask.insert(0, 1)
                attention_mask.append(1)
                all_attention_mask.append(attention_mask)
                word_ids = word_ids_orig[start:end]
                word_ids.insert(0, None)
                word_ids.append(None)
                all_word_ids.append(word_ids)
            else:
                # last sliding window. We have to pad the end in order to have equal length along all windows.
                tokens = tokens_orig[start:end]
                tokens.insert(0, tokenizer.cls_token)
                pads = [tokenizer.pad_token for _ in range(pad_last + 1)]
                tokens.extend(pads)
                all_tokens.append(tokens)
                input_ids = input_ids_orig[start:end]
                input_ids.insert(0, 101)
                pad_ids = [0 for _ in range(pad_last + 1)]
                input_ids.extend(pad_ids)
                all_input_ids.append(input_ids)
                token_type_ids = token_type_ids_orig[start:end]
                token_type_ids.insert(0, 0)
                pad_ids = [0 for _ in range(pad_last + 1)]
                token_type_ids.extend(pad_ids)
                all_token_type_ids.append(token_type_ids)
                attention_mask = attention_mask_orig[start:end]
                attention_mask.insert(0, 1)
                pad_ids = [1 for _ in range(pad_last + 1)]
                attention_mask.extend(pad_ids)
                all_attention_mask.append(attention_mask)
                word_ids = word_ids_orig[start:end]
                word_ids.insert(0, None)
                pad_none = [None for _ in range(pad_last + 1)]
                word_ids.extend(pad_none)
                all_word_ids.append(word_ids)

    slided_tokenized_inputs: Dict[str, Union[List[Union[str, int]], torch.Tensor]] = {}
    if return_tensors == "pt":
        slided_tokenized_inputs["overflow_to_sample_mapping"] = torch.tensor(overflow_to_sample_mapping)
        slided_tokenized_inputs["input_ids"] = torch.tensor(all_input_ids)
        slided_tokenized_inputs["token_type_ids"] = torch.tensor(all_token_type_ids)
        slided_tokenized_inputs["attention_mask"] = torch.tensor(all_attention_mask)
    else:
        slided_tokenized_inputs["overflow_to_sample_mapping"] = overflow_to_sample_mapping  # type: ignore
        slided_tokenized_inputs["input_ids"] = all_input_ids
        slided_tokenized_inputs["token_type_ids"] = all_token_type_ids
        slided_tokenized_inputs["attention_mask"] = all_attention_mask
    slided_tokenized_inputs["word_ids"] = all_word_ids
    slided_tokenized_inputs["tokens"] = all_tokens
    return slided_tokenized_inputs


def raw_features_to_layoutlm_features(
    raw_features: Union[RawLayoutLMFeatures, List[RawLayoutLMFeatures]],
    tokenizer: "PreTrainedTokenizerFast",
    padding: Literal["max_length", "do_not_pad", "longest"] = "max_length",
    truncation: bool = True,
    return_overflowing_tokens: bool = False,
    return_tensors: Optional[Literal["pt"]] = None,
    remove_columns_for_training: bool = False,
    sliding_window_stride: int = 0,
) -> LayoutLMFeatures:
    """
    Mapping raw features to tokenized input sequences for LayoutLM models.

    :param raw_features: A dictionary with the following arguments: `image_id, width, height, ann_ids, words,
                         boxes, dataset_type`.
    :param tokenizer: A fast tokenizer for the model. Note, that the conventional python based tokenizer provided by the
                      Transformer library do not return essential word_id/token_id mappings making the feature
                      generation a lot more difficult. We therefore do not allow these tokenizer.
    :param padding: A padding strategy to be passed to the tokenizer. Must bei either `max_length, longest` or
                    `do_not_pad`.
    :param truncation: If "True" will truncate to a maximum length specified with the argument max_length or to the
                       maximum acceptable input length for the model if that argument is not provided. This will
                       truncate token by token, removing a token from the longest sequence in the pair if a pair of
                       sequences (or a batch of pairs) is provided.
                       If `False` then no truncation (i.e., can output batch with sequence lengths greater than the
                       model maximum admissible input size).
    :param return_overflowing_tokens: If a sequence (due to a truncation strategy) overflows the overflowing tokens can
                                  be returned as an additional batch element. Not that in this case, the number of input
                                  batch samples will be smaller than the output batch samples.
    :param return_tensors: If `pt` will return torch Tensors. If no argument is provided that the batches will be lists
                           of lists.
    :param remove_columns_for_training: Will remove all superfluous columns that are not required for training.
    :param sliding_window_stride: If the output of the tokenizer exceeds the max_length sequence length sliding windows
                                  will be created with each window having max_length sequence input. When using
                                  `sliding_window_stride=0` no strides will be created, otherwise it will create slides
                                  with windows shifted `sliding_window_stride` to the right.
    :return: dictionary with the following arguments:  `image_ids, width, height, ann_ids, input_ids,
             token_type_ids, attention_mask, bbox, labels`.
    """

    if isinstance(raw_features, dict):
        raw_features = [raw_features]

    _has_token_labels = (
        raw_features[0]["dataset_type"] == DatasetType.token_classification
        and raw_features[0].get("labels") is not None
    )
    _has_sequence_labels = (
        raw_features[0]["dataset_type"] == DatasetType.sequence_classification
        and raw_features[0].get("labels") is not None
    )
    _has_labels = bool(_has_token_labels or _has_sequence_labels)
    _image_key = "pixel_values" if "pixel_values" in raw_features[0] else "image"

    if sliding_window_stride:
        return_overflowing_tokens = True
        tokenized_inputs = _tokenize_with_sliding_window(raw_features, tokenizer, sliding_window_stride, return_tensors)

    else:
        tokenized_inputs = tokenizer(
            [dp["words"] for dp in raw_features],
            padding=padding,
            truncation=truncation,
            return_overflowing_tokens=return_overflowing_tokens,
            return_token_type_ids=True,
            is_split_into_words=True,
            return_tensors=return_tensors,
        )

    image_ids = []
    images = []
    widths = []
    heights = []

    token_boxes = []
    token_labels = []
    sequence_labels = []
    token_ann_ids = []
    tokens = []

    for batch_index in range(len(tokenized_inputs["input_ids"])):
        batch_index_orig = batch_index
        if return_overflowing_tokens:
            # we might get more batches when we allow to get returned overflowing tokens
            batch_index_orig = tokenized_inputs["overflow_to_sample_mapping"][batch_index]

        if _image_key in raw_features[batch_index_orig]:
            images.append(raw_features[batch_index_orig][_image_key])
        image_ids.append(raw_features[batch_index_orig]["image_id"])
        widths.append(raw_features[batch_index_orig]["width"])
        heights.append(raw_features[batch_index_orig]["height"])

        ann_ids = raw_features[batch_index_orig]["ann_ids"]
        boxes = raw_features[batch_index_orig]["bbox"]
        if _has_token_labels:
            labels = raw_features[batch_index_orig]["labels"]

        if isinstance(tokenized_inputs, dict):
            word_ids = tokenized_inputs["word_ids"][batch_index]
            token_batch = tokenized_inputs["tokens"][batch_index]
        else:
            word_ids = tokenized_inputs.word_ids(batch_index=batch_index)
            token_batch = tokenized_inputs.tokens(batch_index=batch_index)

        token_batch_ann_ids = []
        token_batch_boxes = []
        token_batch_labels = []
        for idx, word_id in enumerate(word_ids):
            # Special tokens have a word id that is None. We make a lookup for the specific token and append a dummy
            # bounding box accordingly
            if word_id is None:
                if token_batch[idx] == tokenizer.cls_token:
                    token_batch_boxes.append(_CLS_BOX)
                    token_batch_ann_ids.append(tokenizer.cls_token)
                elif token_batch[idx] in (tokenizer.sep_token, tokenizer.pad_token):
                    token_batch_boxes.append(_SEP_BOX)
                    if token_batch[idx] == tokenizer.sep_token:
                        token_batch_ann_ids.append(tokenizer.sep_token)
                    else:
                        token_batch_ann_ids.append(tokenizer.pad_token)
                else:
                    raise ValueError(f"Special token {token_batch[idx]} not allowed")
                if _has_token_labels:
                    token_batch_labels.append(-100)
            else:
                token_batch_boxes.append(boxes[word_id])
                token_batch_ann_ids.append(ann_ids[word_id])
                if _has_token_labels:
                    token_batch_labels.append(labels[word_id])

        token_labels.append(token_batch_labels)
        token_boxes.append(token_batch_boxes)
        token_ann_ids.append(token_batch_ann_ids)
        tokens.append(token_batch)
        if _has_sequence_labels:
            sequence_labels.append(raw_features[batch_index_orig]["labels"][0])

    input_dict = {
        "image_ids": image_ids,
        "width": widths,
        "height": heights,
        "ann_ids": token_ann_ids,
        "input_ids": tokenized_inputs["input_ids"],
        "token_type_ids": tokenized_inputs["token_type_ids"],
        "attention_mask": tokenized_inputs["attention_mask"],
        "bbox": token_boxes,
        "tokens": tokens,
    }

    # will only add the image to features if it has been passed as raw feature
    if images:
        input_dict[_image_key] = images

    if _has_labels:
        input_dict["labels"] = token_labels if _has_token_labels else sequence_labels

    if remove_columns_for_training:
        input_dict.pop("image_ids")
        input_dict.pop("width")
        input_dict.pop("height")
        input_dict.pop("ann_ids")
        input_dict.pop("tokens")

    if return_tensors == "pt":
        return features_to_pt_tensors(LayoutLMFeatures(input_dict))
    return LayoutLMFeatures(input_dict)


@dataclass
class LayoutLMDataCollator:
    """
    Data collator that will dynamically tokenize, pad and truncate the inputs received.

    :param tokenizer: A fast tokenizer for the model. Note, that the conventional python based tokenizer provided by the
                      Transformer library do not return essential word_id/token_id mappings making the feature
                      generation a lot more difficult. We therefore do not allow these tokenizer.
    :param padding: A padding strategy to be passed to the tokenizer. Must bei either `max_length, longest` or
                    `do_not_pad`.
    :param truncation: If "True" will truncate to a maximum length specified with the argument max_length or to the
                       maximum acceptable input length for the model if that argument is not provided. This will
                       truncate token by token, removing a token from the longest sequence in the pair if a pair of
                       sequences (or a batch of pairs) is provided.
                       If `False` then no truncation (i.e., can output batch with sequence lengths greater than the
                       model maximum admissible input size).
    :param return_overflowing_tokens: If a sequence (due to a truncation strategy) overflows the overflowing tokens can
                                  be returned as an additional batch element. Not that in this case, the number of input
                                  batch samples will be smaller than the output batch samples.
    :param return_tensors: If `pt` will return torch Tensors. If no argument is provided that the batches will be lists
                           of lists.
    """

    tokenizer: "PreTrainedTokenizerFast"
    padding: Literal["max_length", "do_not_pad", "longest"] = field(default="max_length")
    truncation: bool = field(default=True)
    return_overflowing_tokens: bool = field(default=False)
    return_tensors: Optional[Literal["pt"]] = field(default=None)
    sliding_window_stride: int = field(default=0)

    def __post_init__(self) -> None:
        assert isinstance(self.tokenizer, PreTrainedTokenizerFast), "Tokenizer must be a fast tokenizer"
        if self.return_tensors:
            assert self.padding not in ("do_not_pad",), self.padding
            assert self.truncation, self.truncation
        if self.return_overflowing_tokens:
            assert self.truncation, self.truncation

    def __call__(self, raw_features: Union[RawLayoutLMFeatures, List[RawLayoutLMFeatures]]) -> LayoutLMFeatures:
        """
        Calling the DataCollator to form model inputs for training and inference. Takes a single raw
        :param raw_features: A dictionary with the following arguments: `image_id, width, height, ann_ids, words,
                             boxes, dataset_type`.
        :return: LayoutLMFeatures with arguments `image_ids, width, height, ann_ids, input_ids,
                 token_type_ids, attention_masks, boxes, labels`.
        """
        return raw_features_to_layoutlm_features(
            raw_features,
            self.tokenizer,
            self.padding,
            self.truncation,
            self.return_overflowing_tokens,
            self.return_tensors,
            True,
            self.sliding_window_stride,
        )


@curry
def image_to_layoutlm_features(
    dp: Image,
    tokenizer: "PreTrainedTokenizerFast",
    padding: Literal["max_length", "do_not_pad", "longest"] = "max_length",
    truncation: bool = True,
    return_overflowing_tokens: bool = False,
    return_tensors: Optional[Literal["pt"]] = "pt",
    input_width: int = 1000,
    input_height: int = 1000,
    image_width: int = 1000,
    image_height: int = 1000,
    color_mode: Literal["BGR", "RGB"] = "BGR",
    pixel_mean: Optional[npt.NDArray[np.float32]] = None,
    pixel_std: Optional[npt.NDArray[np.float32]] = None,
    segment_positions: Optional[Union[LayoutType, Sequence[LayoutType]]] = None,
    sliding_window_stride: int = 0,
) -> Optional[LayoutLMFeatures]:
    """
    Mapping function to generate layoutlm features from `Image` to be used for inference in a pipeline component.
    `LanguageModelPipelineComponent` has a positional argument `mapping_to_lm_input_func` that must be chosen
    with respect to the language model chosen. This mapper is devoted to generating features for LayoutLM. It will be
    used internally in `LMTokenClassifierService`.

            tokenizer = LayoutLMTokenizer.from_pretrained("mrm8488/layoutlm-finetuned-funsd")
            layoutlm = HFLayoutLmTokenClassifier("path/to/config.json","path/to/model.bin",
                                                  categories_explicit=['B-ANSWER', 'B-QUESTION', 'O'])

            layoutlm_service = LMTokenClassifierService(tokenizer,layoutlm)

    :param dp: Image datapoint
    :param tokenizer: Tokenizer compatible with the language model
    :param padding: A padding strategy to be passed to the tokenizer. Must bei either `max_length, longest` or
                    `do_not_pad`.
    :param truncation: If "True" will truncate to a maximum length specified with the argument max_length or to the
                       maximum acceptable input length for the model if that argument is not provided. This will
                       truncate token by token, removing a token from the longest sequence in the pair if a pair of
                       sequences (or a batch of pairs) is provided.
                       If `False` then no truncation (i.e., can output batch with sequence lengths greater than the
                       model maximum admissible input size).
    :param return_overflowing_tokens: If a sequence (due to a truncation strategy) overflows the overflowing tokens
                                      can be returned as an additional batch element. Not that in this case, the number
                                      of input batch samples will be smaller than the output batch samples.
    :param return_tensors: Output tensor features. Either 'pt' for PyTorch models or None, if features should be
                           returned in list objects.
    :param input_width: Standard input size for image coordinates. All LayoutLM models require input features to be
                        normalized to an image width equal to 1000.
    :param input_height: Standard input size for image coordinates. All LayoutLM models require input features to be
                         normalized to an image height equal to 1000.
    :param image_width: Some models (e.g. `Layoutlmv2`) assume box coordinates to be normalized to input_width, whereas
                        the image has to be resized to a different width. This input will only resize the `image` width.
    :param image_height: Some models (e.g. `Layoutlmv2`) assume box coordinates to be normalized to input_height,
                         whereas the image has to be resized to a different height. This input will only resize the
                         `image` height.
    :param color_mode: Either "BGR" or "RGB". Note, that LayoutLMv2 uses "BGR" because of Detectron2 backbone, whereas
                       LayoutLMv3 uses "RGB".
    :param pixel_mean: (3,) array for "BGR" resp. "RGB" mean
    :param pixel_std: (3,) array for "BGR" resp. "RGB" std
    :param segment_positions: Using bounding boxes of segment instead of words improves model accuracy significantly.
                              Choose a single or a sequence of layout segments to use their bounding boxes. Note, that
                              the layout segments need to have a child-relationship with words. If a word does not
                              appear as child, it will use the word bounding box.
    :param sliding_window_stride: If the output of the tokenizer exceeds the max_length sequence length a sliding
                                  windows will be created with each window having max_length sequence input. When using
                                  `sliding_window_stride=0` no strides will be created, otherwise it will create slides
                                  with windows shifted `sliding_window_stride` to the right.
    :return: A dict of layoutlm features
    """
    raw_features = image_to_raw_layoutlm_features(
        None,
        input_width,
        input_height,
        image_width,
        image_height,
        color_mode,
        pixel_mean,
        pixel_std,
        True,
        segment_positions,
    )(dp)
    if raw_features is None:
        return None
    features = raw_features_to_layoutlm_features(
        raw_features,
        tokenizer,
        padding,
        truncation,
        return_overflowing_tokens,
        return_tensors=return_tensors,
        sliding_window_stride=sliding_window_stride,
    )
    return features



# -*- coding: utf-8 -*-
# File: convert.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Conversion functions associated to functionalities of datapoint classes
"""
import base64
import copy
from dataclasses import fields, is_dataclass
from io import BytesIO
from shutil import which
from typing import Any, Optional, Union

import cv2
import numpy as np
from numpy import uint8
from numpy.typing import NDArray
from PyPDF2 import PdfFileReader

from ..utils.detection_types import ImageType
from ..utils.develop import deprecated
from ..utils.pdf_utils import pdf_to_np_array

__all__ = [
    "convert_b64_to_np_array",
    "convert_np_array_to_b64",
    "convert_np_array_to_b64_b",
    "convert_pdf_bytes_to_np_array_v2",
    "box_to_point4",
    "point4_to_box",
    "as_dict",
]


def as_dict(obj: Any, dict_factory) -> Union[Any]:  # type: ignore
    """
    custom func: as_dict to use instead of `dataclasses.asdict` . It also checks if a dataclass has a
    'remove_keys' and will remove all attributes that are returned. Ensures that private attributes are not taken
    into account when generating a dict.

    :param obj: Object to convert into a dict.
    :param dict_factory: A factory to generate the dict.
    """

    if is_dataclass(obj):
        result = []
        for attribute in fields(obj):
            value = as_dict(getattr(obj, attribute.name), dict_factory)
            if hasattr(obj, "remove_keys"):
                if attribute.name in obj.remove_keys():
                    continue
            result.append((attribute.name, value))
        return dict_factory(result)
    if isinstance(obj, (list, tuple)):
        return type(obj)(as_dict(v, dict_factory) for v in obj)  # pylint:disable=E0110
    if isinstance(obj, dict):
        return type(obj)((as_dict(k, dict_factory), as_dict(v, dict_factory)) for k, v in obj.items())
    if isinstance(obj, (np.float32, np.float64)):
        obj = obj.astype(float)
    return copy.deepcopy(obj)


def convert_b64_to_np_array(image: str) -> ImageType:
    """
    Converts an image in base4 string encoding representation to a numpy array of shape (width,height,channel).

    :param image: An image as base64 string.
    :return: numpy array.
    """
    np_array = np.fromstring(base64.b64decode(image), np.uint8)  # type: ignore
    np_array = cv2.imdecode(np_array, cv2.IMREAD_COLOR).astype(np.float32)
    return np_array.astype(uint8)


def convert_np_array_to_b64(np_image: ImageType) -> str:
    """
    Converts an image from numpy array into a base64 string encoding representation

    :param np_image: An image as numpy array.
    :return: An image as base64 string.
    """
    np_encode = cv2.imencode(".png", np_image)
    image = base64.b64encode(np_encode[1]).decode("utf-8")
    return image


def convert_np_array_to_b64_b(np_image: ImageType) -> bytes:
    """
    Converts an image from numpy array into a base64 bytes encoding representation

    :param np_image: An image as numpy array.
    :return: An image as base64 bytes.
    """
    np_encode = cv2.imencode(".png", np_image)
    b_image = np_encode[1].tobytes()
    return b_image


@deprecated("Use convert_pdf_bytes_to_np_array_v2", "2022-02-23")
def convert_pdf_bytes_to_np_array(pdf_bytes: bytes, dpi: Optional[int] = None) -> ImageType:
    """
    Converts a pdf passed as bytes into a numpy array. Note, that this method expects poppler to be installed.
    Please check the installation guides at https://poppler.freedesktop.org/ . If no value for dpi is provided
    the output size will be determined by the mediaBox of the pdf file ready. Note, that the image size will be in
    this case rather small.

    :param pdf_bytes: A pdf as bytes object. A byte representation can from a pdf file can be generated e.g. with
                      `utils.fs.load_bytes_from_pdf_file`
    :param dpi: The dpi value of the resulting output image. For high resolution set dpi=300.
    :return: Image as numpy array.
    """
    from pdf2image import convert_from_bytes  # type: ignore # pylint: disable=C0415, E0401

    assert which("pdftoppm") is not None, "convert_pdf_bytes_to_np_array requires poppler to be installed"

    with BytesIO(pdf_bytes) as pdf_file:
        pdf = PdfFileReader(pdf_file).getPage(0)
    shape = pdf.mediaBox
    height = shape[3] - shape[1]
    width = shape[2] - shape[0]
    buffered = BytesIO()

    if dpi is None:

        image = convert_from_bytes(pdf_bytes, size=(width, height))[0]
    else:
        image = convert_from_bytes(pdf_bytes, dpi=dpi)[0]

    image.save(buffered, format="JPEG")
    image = base64.b64encode(buffered.getvalue()).decode("utf-8")
    np_array = convert_b64_to_np_array(image)
    return np_array.astype(uint8)


def convert_pdf_bytes_to_np_array_v2(pdf_bytes: bytes, dpi: Optional[int] = None) -> ImageType:
    """
    Converts a pdf passed as bytes into a numpy array. Note, that this method expects poppler to be installed. This
    function, however does not rely on the wrapper pdf2image but uses a function of this lib which calls poppler
    directly.

    :param pdf_bytes: A pdf as bytes object. A byte representation can from a pdf file can be generated e.g. with
                      `utils.fs.load_bytes_from_pdf_file`
    :param dpi: The dpi value of the resulting output image. For high resolution set dpi=300.
    :return: Image as numpy array.
    """

    with BytesIO(pdf_bytes) as pdf_file:
        pdf = PdfFileReader(pdf_file).getPage(0)
    shape = pdf.mediaBox
    height = shape[3] - shape[1]
    width = shape[2] - shape[0]

    if dpi is None:
        return pdf_to_np_array(pdf_bytes, size=(int(width), int(height)))
    return pdf_to_np_array(pdf_bytes, dpi=dpi)


def box_to_point4(boxes: NDArray[np.float32]) -> NDArray[np.float32]:
    """
    :param boxes: nx4
    :return: (nx4)x2
    """
    box = boxes[:, [0, 1, 2, 3, 0, 3, 2, 1]]
    box = box.reshape((-1, 2))
    return box


def point4_to_box(points: NDArray[np.float32]) -> NDArray[np.float32]:
    """
    :param points: (nx4)x2
    :return: nx4 boxes (x1y1x2y2)
    """
    points = points.reshape((-1, 4, 2))
    min_xy = points.min(axis=1)  # nx2
    max_xy = points.max(axis=1)  # nx2
    return np.concatenate((min_xy, max_xy), axis=1)



# -*- coding: utf-8 -*-
# File: base.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for the base class for evaluations and metrics
"""

from abc import ABC, abstractmethod
from typing import Any, Callable, List, Optional, Tuple

from ..dataflow import DataFlow
from ..datasets.info import DatasetCategories
from ..utils.detection_types import JsonDict
from ..utils.file_utils import Requirement


class MetricBase(ABC):
    """
    Base class for metrics. Metrics only exist as classes and are not instantiated. They consist of two class variables:

        - A metric function that reads ground truth and prediction in an already transformed data format and
          returns a distance as an evaluation result.

        - A mapping function that transforms an image datapoint into a valid input format ready to ingest by the metric
          function.

    Using `get_distance`, ground truth and prediction dataflow can be read in and evaluated.
    `dump` is a helper method that is often called via `get_distance`. Here, the dataflows should be
    executed and the results should be saved in separate lists.
    """

    name: str
    metric: Callable[[Any, Any], Optional[Any]]
    mapper: Callable[[Any, Any], Optional[Any]]
    _results: List[JsonDict]

    def __new__(cls, *args, **kwargs):  # type: ignore # pylint: disable=W0613
        requirements = cls.get_requirements()
        name = cls.__name__ if hasattr(cls, "__name__") else cls.__class__.__name__
        if not all(requirement[1] for requirement in requirements):
            raise ImportError(
                "\n".join(
                    [f"{name} has the following dependencies:"]
                    + [requirement[2] for requirement in requirements if not requirement[1]]
                )
            )
        return super().__new__(cls)

    @classmethod
    @abstractmethod
    def get_requirements(cls) -> List[Requirement]:
        """
        Get a list of requirements for running the detector
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def get_distance(
        cls, dataflow_gt: DataFlow, dataflow_predictions: DataFlow, categories: DatasetCategories
    ) -> List[JsonDict]:
        """
        Takes of the ground truth processing strand as well as the prediction strand and generates the metric results.

        :param dataflow_gt: Dataflow with ground truth annotations.
        :param dataflow_predictions: Dataflow with predictions.
        :param categories:  DatasetCategories with respect to the underlying dataset.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def dump(
        cls, dataflow_gt: DataFlow, dataflow_predictions: DataFlow, categories: DatasetCategories
    ) -> Tuple[Any, Any]:
        """
        Dump the dataflow with ground truth annotations and predictions. Use it as auxiliary method and call it from
        `get_distance`.

        :param dataflow_gt: Dataflow with ground truth annotations.
        :param dataflow_predictions: Dataflow with predictions.
        :param categories: DatasetCategories with respect to the underlying dataset.
        """
        raise NotImplementedError

    @classmethod
    def result_list_to_dict(cls, results: List[JsonDict]) -> JsonDict:
        """
        Converts the result from `get_distance` to a dict. It concatenates all keys of the inner dict and uses
        the metric result 'val' as value.

        :param results: List of dict as input
        :return: Dict with metric results.
        """
        output: JsonDict = {}
        for res in results:
            new_key = ""
            new_val = 0.0
            for k, val in res.items():

                if str(val) != "":
                    if k != "val":
                        if k != "key":
                            new_key += k + "/" + str(val) + "/"
                        else:
                            new_key += str(val) + "/"
                    else:
                        new_val = val
            output[new_key[:-1]] = new_val

        return output

    @classmethod
    def print_result(cls) -> None:
        """Print metric result. Overwrite this method if you want a specific output"""



# -*- coding: utf-8 -*-
# File: transform.py

# Copyright 2022 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for deterministic image transformations and the sometimes necessary recalculation
of coordinates. Most have the ideas have been taken from
<https://github.com/tensorpack/dataflow/blob/master/dataflow/dataflow/imgaug/transform.py> .
"""

from abc import ABC, abstractmethod
from typing import Literal, Optional, Union

import cv2
import numpy as np
import numpy.typing as npt
from numpy import float32

from .detection_types import ImageType

__all__ = ["ResizeTransform", "InferenceResize", "PadTransform"]


class BaseTransform(ABC):
    """
    A deterministic image transformation. This class is also the place to provide a default implementation to any
    `apply_xxx` method. The current default is to raise NotImplementedError in any such methods.
    All subclasses should implement `apply_image`. The image should be of type uint8 in range [0, 255], or
    floating point images in range [0, 1] or [0, 255]. Some subclasses may implement `apply_coords`, when applicable.
    It should take and return a numpy array of Nx2, where each row is the (x, y) coordinate.
    The implementation of each method may choose to modify its input data in-place for efficient transformation.
    """

    @abstractmethod
    def apply_image(self, img: ImageType) -> ImageType:
        """The transformation that should be applied to the image"""
        raise NotImplementedError


class ResizeTransform(BaseTransform):
    """
    Resize the image.
    """

    def __init__(
        self,
        h: Union[int, float],
        w: Union[int, float],
        new_h: Union[int, float],
        new_w: Union[int, float],
        interp: str,
    ):
        """
        :param h: height
        :param w: width
        :param new_h: target height
        :param new_w: target width
        :param interp: cv2 interpolation method like cv2.INTER_NEAREST, cv2.INTER_LINEAR,
                       cv2.INTER_AREA
        """
        self.h = h
        self.w = w
        self.new_h = new_h
        self.new_w = new_w
        self.interp = interp

    def apply_image(self, img: ImageType) -> ImageType:
        assert img.shape[:2] == (self.h, self.w)
        ret = cv2.resize(img, (self.new_w, self.new_h), interpolation=self.interp)
        if img.ndim == 3 and ret.ndim == 2:
            ret = ret[:, :, np.newaxis]
        return ret

    def apply_coords(self, coords: npt.NDArray[float32]) -> npt.NDArray[float32]:
        """Transformation that should be applied to coordinates"""
        coords[:, 0] = coords[:, 0] * (self.new_w * 1.0 / self.w)
        coords[:, 1] = coords[:, 1] * (self.new_h * 1.0 / self.h)
        return coords


class InferenceResize:
    """
    Try resizing the shortest edge to a certain number while avoiding the longest edge to exceed max_size. This is
    the inference version of `extern.tp.frcnn.common.CustomResize` .
    """

    def __init__(self, short_edge_length: int, max_size: int, interp: str = cv2.INTER_LINEAR) -> None:
        """
        :param short_edge_length: a [min, max] interval from which to sample the shortest edge length.
        :param max_size: maximum allowed longest edge length.
        """
        self.short_edge_length = short_edge_length
        self.max_size = max_size
        self.interp = interp

    def get_transform(self, img: ImageType) -> ResizeTransform:
        """
        get transform
        """
        h, w = img.shape[:2]
        new_w: Union[int, float]
        new_h: Union[int, float]

        scale = self.short_edge_length * 1.0 / min(h, w)

        if h < w:
            new_h, new_w = self.short_edge_length, scale * w
        else:
            new_h, new_w = scale * h, self.short_edge_length
        if max(new_h, new_w) > self.max_size:
            scale = self.max_size * 1.0 / max(new_h, new_w)
            new_h = new_h * scale
            new_w = new_w * scale
        new_w = int(new_w + 0.5)
        new_h = int(new_h + 0.5)
        return ResizeTransform(h, w, new_h, new_w, self.interp)


def normalize_image(image: ImageType, pixel_mean: npt.NDArray[float32], pixel_std: npt.NDArray[float32]) -> ImageType:
    """
    Preprocess pixel values of an image by rescaling.

    :param image: image as np.array
    :param pixel_mean: (3,) array
    :param pixel_std: (3,) array
    """
    return (image - pixel_mean) * (1.0 / pixel_std)


def pad_image(image: ImageType, top: int, right: int, bottom: int, left: int) -> ImageType:
    """Pad an image with white color and with given top/bottom/right/left pixel values. Only white padding is
    currently supported

    :param image: image as np.array
    :param top: Top pixel value to pad
    :param right: Right pixel value to pad
    :param bottom: Bottom pixel value to pad
    :param left: Left pixel value to pad
    """
    return np.pad(image, ((left, right), (top, bottom), (0, 0)), "constant", constant_values=(255))


class PadTransform(BaseTransform):
    """
    A transform for padding images left/right/top/bottom-wise.
    """

    def __init__(
        self,
        top: int,
        right: int,
        bottom: int,
        left: int,
        mode: Literal["xyxy", "xywh"] = "xyxy",
    ):
        """
        :param top: padding top image side
        :param right: padding right image side
        :param bottom: padding bottom image side
        :param left: padding left image side
        :param mode: bounding box mode. Needed for transforming coordinates.
        """
        self.top = top
        self.right = right
        self.bottom = bottom
        self.left = left
        self.image_width: Optional[int] = None
        self.image_height: Optional[int] = None
        self.mode = mode

    def apply_image(self, img: ImageType) -> ImageType:
        """Apply padding to image"""
        self.image_width = img.shape[1]
        self.image_height = img.shape[0]
        return pad_image(img, self.top, self.right, self.bottom, self.left)

    def apply_coords(self, coords: npt.NDArray[float32]) -> npt.NDArray[float32]:
        """Transformation that should be applied to coordinates"""
        if self.mode == "xyxy":
            coords[:, 0] = coords[:, 0] + self.left
            coords[:, 1] = coords[:, 1] + self.top
            coords[:, 2] = coords[:, 2] + self.left
            coords[:, 3] = coords[:, 3] + self.top
        else:
            coords[:, 0] = coords[:, 0] + self.left
            coords[:, 1] = coords[:, 1] + self.top
        return coords

    def inverse_apply_coords(self, coords: npt.NDArray[float32]) -> npt.NDArray[float32]:
        """Inverse transformation going back from coordinates of padded image to original image"""
        if self.image_height is None or self.image_width is None:
            raise ValueError("Initialize image_width and image_height first")

        if self.mode == "xyxy":
            coords[:, 0] = np.maximum(coords[:, 0] - self.left, np.zeros(coords[:, 0].shape))
            coords[:, 1] = np.maximum(coords[:, 1] - self.top, np.zeros(coords[:, 1].shape))
            coords[:, 2] = np.minimum(coords[:, 2] - self.left, np.ones(coords[:, 2].shape) * self.image_width)
            coords[:, 3] = np.minimum(coords[:, 3] - self.top, np.ones(coords[:, 3].shape) * self.image_height)
        else:
            coords[:, 0] = np.maximum(coords[:, 0] - self.left, np.zeros(coords[:, 0].shape))
            coords[:, 1] = np.maximum(coords[:, 1] - self.top, np.zeros(coords[:, 1].shape))
        return coords

    def clone(self) -> "PadTransform":
        """clone"""
        return self.__class__(self.top, self.right, self.bottom, self.left, self.mode)



# -*- coding: utf-8 -*-
# File: xfund.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for XFUND dataset. Install the dataset following the folder structure

    xfund
    ├── de_train
    │ ├── de_train_0.jpg
    │ ├── de_train_1.jpg
    ├── de.train.json
    ├── de_val
    │ ├── de_val_0.jpg
    ├── es_train
"""

import json
import os
from typing import Mapping, Sequence, Union

from ...dataflow import CustomDataFromList, DataFlow, MapData
from ...datasets.info import DatasetInfo
from ...mapper.cats import cat_to_sub_cat, filter_cat
from ...mapper.xfundstruct import xfund_to_image
from ...utils.detection_types import JsonDict
from ...utils.settings import BioTag, DatasetType, LayoutType, ObjectTypes, TokenClasses, TokenClassWithTag, WordType
from ..base import _BuiltInDataset
from ..dataflow_builder import DataFlowBaseBuilder
from ..info import DatasetCategories
from ..registry import dataset_registry

_NAME = "xfund"
_DESCRIPTION = (
    "XFUND is a multilingual form understanding benchmark dataset that includes human-labeled forms with \n"
    "key-value pairs in 7 languages (Chinese, Japanese, Spanish, French, Italian, German, Portuguese)."
)
_LICENSE = (
    "The content of this project itself is licensed under the Attribution-NonCommercial-ShareAlike 4.0 \n"
    "International (CC BY-NC-SA 4.0) Portions of the source code are based on the transformers project. \n"
    "Microsoft Open Source Code of Conduct"
)
_URL = "https://github.com/doc-analysis/XFUND/releases/tag/v1.0"
_SPLITS: Mapping[str, str] = {"train": "train", "val": "val"}
_TYPE = DatasetType.token_classification
_LOCATION = "xfund"
_ANNOTATION_FILES: Mapping[str, Union[str, Sequence[str]]] = {
    "train": [
        "de.train.json",
        "es.train.json",
        "fr.train.json",
        "it.train.json",
        "ja.train.json",
        "pt.train.json",
        "zh.train.json",
    ],
    "val": ["de.val.json", "es.val.json", "fr.val.json", "it.val.json", "ja.val.json", "pt.val.json", "zh.val.json"],
}
_INIT_CATEGORIES = [LayoutType.word, LayoutType.text]
_SUB_CATEGORIES: Mapping[ObjectTypes, Mapping[ObjectTypes, Sequence[ObjectTypes]]]
_SUB_CATEGORIES = {
    LayoutType.word: {
        WordType.token_class: [TokenClasses.other, TokenClasses.question, TokenClasses.answer, TokenClasses.header],
        WordType.tag: [BioTag.inside, BioTag.outside, BioTag.begin],
        WordType.token_tag: [
            TokenClassWithTag.b_answer,
            TokenClassWithTag.b_header,
            TokenClassWithTag.b_question,
            TokenClassWithTag.i_answer,
            TokenClassWithTag.i_header,
            TokenClassWithTag.i_question,
            BioTag.outside,
        ],
    },
    LayoutType.text: {
        WordType.token_class: [TokenClasses.other, TokenClasses.question, TokenClasses.answer, TokenClasses.header]
    },
}

_LANGUAGES = ["de", "es", "fr", "it", "ja", "pt", "zh"]


@dataset_registry.register("xfund")
class Xfund(_BuiltInDataset):
    """
    Xfund
    """

    _name = _NAME

    @classmethod
    def _info(cls) -> DatasetInfo:
        return DatasetInfo(name=_NAME, description=_DESCRIPTION, license=_LICENSE, url=_URL, splits=_SPLITS, type=_TYPE)

    def _categories(self) -> DatasetCategories:
        return DatasetCategories(init_categories=_INIT_CATEGORIES, init_sub_categories=_SUB_CATEGORIES)

    def _builder(self) -> "XfundBuilder":
        return XfundBuilder(location=_LOCATION, annotation_files=_ANNOTATION_FILES)


class XfundBuilder(DataFlowBaseBuilder):
    """
    Xfund dataflow builder
    """

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Returns a dataflow from which you can stream datapoints of images. The following arguments affect the returns
        of the dataflow:

        `split:` Split of the dataset. `train` and `val` is available

        `load_image:` Will load the image for each datapoint.  Default: `False`

        `max_datapoints:` Will stop iterating after max_datapoints. Default: `None`

        `languages:` Will select only samples of selected languages. Available languages: `de`, `es`, `fr`, `it`, `ja` ,
                     `pt`, `zh`. If default will take any language.
        :return: Dataflow
        """

        split = str(kwargs.get("split", "val"))
        load_image = kwargs.get("load_image", False)
        max_datapoints = kwargs.get("max_datapoints")
        language = kwargs.get("languages")

        if max_datapoints is not None:
            max_datapoints = int(max_datapoints)

        if language is None:
            languages = _LANGUAGES
        elif isinstance(language, str):
            languages = [language]
        else:
            raise TypeError("language requires to be a string")

        if not all(elem in _LANGUAGES for elem in languages):
            raise ValueError("Not all languages available")

        # Load
        df: DataFlow
        path_ann_files = [
            os.path.join(self.get_workdir(), ann_file)
            for ann_file in self.annotation_files[split]
            if ann_file.split(".")[0] in languages
        ]

        datapoints = []
        for path_ann in path_ann_files:
            with open(path_ann, "r", encoding="utf-8") as file:
                anns = json.loads(file.read())
                datapoints.extend(anns["documents"])
        df = CustomDataFromList(datapoints, max_datapoints=max_datapoints)

        # Map
        def replace_filename(dp: JsonDict) -> JsonDict:
            folder = "_".join(dp["id"].split("_", 2)[:2])
            dp["img"]["fname"] = os.path.join(self.get_workdir(), folder, dp["img"]["fname"])
            return dp

        df = MapData(df, replace_filename)
        categories_name_as_key = self.categories.get_categories(init=True, name_as_key=True)
        token_class_names_mapping = {
            "other": TokenClasses.other,
            "question": TokenClasses.question,
            "answer": TokenClasses.answer,
            "header": TokenClasses.header,
        }
        ner_token_to_id_mapping = self.categories.get_sub_categories(
            categories=LayoutType.word,
            sub_categories={LayoutType.word: [WordType.token_tag, WordType.tag, WordType.token_class]},
            keys=False,
            values_as_dict=True,
            name_as_key=True,
        )
        df = MapData(
            df,
            xfund_to_image(
                load_image, False, categories_name_as_key, token_class_names_mapping, ner_token_to_id_mapping
            ),
        )
        if self.categories.is_filtered():
            df = MapData(
                df,
                filter_cat(  # pylint: disable=E1120
                    self.categories.get_categories(as_dict=False, filtered=True),
                    self.categories.get_categories(as_dict=False, filtered=False),
                ),
            )

        if self.categories.is_cat_to_sub_cat():
            df = MapData(
                df,
                cat_to_sub_cat(self.categories.get_categories(name_as_key=True), self.categories.cat_to_sub_cat),
            )
        return df



# -*- coding: utf-8 -*-
# File: registry.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for DatasetRegistry
"""

import catalogue  # type: ignore
from tabulate import tabulate
from termcolor import colored

from .base import DatasetBase

__all__ = ["dataset_registry", "get_dataset", "print_dataset_infos"]


dataset_registry = catalogue.create("deepdoctection", "datasets", entry_points=True)


def get_dataset(name: str) -> DatasetBase:
    """
    Returns an instance of a dataset with a given name. This instance can be used to customize the dataflow output

    **Example:**

            dataset = get_dataset("some_name")
            dataset.dataflow.categories.filter_categories(["cat1","cat2"])
            df = dataset.dataflow.build(split="train")

            for dp in df:
                # do something

    :param name: A dataset name
    :return: An instance of a dataset
    """
    return dataset_registry.get(name)()


def print_dataset_infos(add_license: bool = True, add_info: bool = True) -> None:
    """
    Prints a table with all registered datasets and some basic information (name, license and optionally description)

    :param add_license: Whether to add the license type of the dataset
    :param add_info: Whether to add a description of the dataset
    """

    data = dataset_registry.get_all()
    num_columns = min(6, len(data))
    infos = []
    for dataset in data.items():
        info = [dataset[0]]
        if add_license:
            info.append(dataset[1]._info().license)  # pylint: disable=W0212
        if add_info:
            info.append(dataset[1]._info().description)  # pylint: disable=W0212
        infos.append(info)
    header = ["dataset"]
    if add_license:
        header.append("license")
    if add_info:
        header.append("description")
    table = tabulate(
        infos, headers=header * (num_columns // 2), tablefmt="fancy_grid", stralign="left", numalign="left"
    )
    print(colored(table, "cyan"))



# -*- coding: utf-8 -*-
# File: fs.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Methods and classes that incorporate filesystem operations as well as file checks
"""

import errno
import json
import os
from base64 import b64encode
from io import BytesIO
from pathlib import Path
from typing import Callable, Literal, Optional, Protocol, Union, overload
from urllib.request import urlretrieve

from cv2 import IMREAD_COLOR, imread

from .detection_types import ImageType, JsonDict, Pathlike
from .logger import logger
from .pdf_utils import get_pdf_file_reader, get_pdf_file_writer
from .tqdm import get_tqdm
from .utils import FileExtensionError, is_file_extension

__all__ = [
    "load_image_from_file",
    "load_bytes_from_pdf_file",
    "get_load_image_func",
    "maybe_path_or_pdf",
    "download",
    "mkdir_p",
    "is_file_extension",
    "load_json",
    "FileExtensionError",
]


def sizeof_fmt(num: float, suffix: str = "B") -> str:
    """
    Converting bytes number into human-readable
    """
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")
def mkdir_p(dir_name: Pathlike) -> None:
    """
    Like "mkdir -p", make a dir recursively, but do nothing if the dir exists

    :param dir_name: name of dir
    """
    assert dir_name is not None
    if dir_name == "" or os.path.isdir(dir_name):
        return None
    try:
        os.makedirs(dir_name)
    except OSError as err:
        if err.errno != errno.EEXIST:
            raise err


# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")
def download(url: str, directory: Pathlike, file_name: Optional[str] = None, expect_size: Optional[int] = None) -> str:
    """
    Download URL to a directory. Will figure out the filename automatically from URL, if not given.
    """
    mkdir_p(directory)
    if file_name is None:
        file_name = url.split("/")[-1]
    f_path = os.path.join(directory, file_name)

    if os.path.isfile(f_path):
        if expect_size is not None and os.stat(f_path).st_size == expect_size:
            logger.info("File %s exists! Skip download.", file_name)
            return f_path
        logger.warning("File %s exists. Will overwrite with a new download!", file_name)
    else:
        logger.info("File %s will be downloaded.", file_name)

    def hook(total):  # type: ignore
        last_b = [0]

        def inner(byte, b_size, t_size=None):  # type: ignore
            if t_size is not None:
                total.total = t_size
            total.update((byte - last_b[0]) * b_size)
            last_b[0] = byte

        return inner

    try:
        with get_tqdm(unit="B", unit_scale=True, miniters=1, desc=file_name) as time:
            f_path, _ = urlretrieve(url, f_path, reporthook=hook(time))
        stat_info = os.stat(f_path)
        size = stat_info.st_size
    except IOError:
        logger.error("Failed to download %s", url)
        raise
    assert size > 0, f"Downloaded an empty file from {url}!"

    if expect_size is not None and size != expect_size:
        logger.error("File downloaded from %s does not match the expected size!", url)
        logger.error("You may have downloaded a broken file, or the upstream may have modified the file.")

    logger.info("Successfully downloaded %s. %s.", file_name, sizeof_fmt(size))
    return f_path


@overload
def load_image_from_file(path: Pathlike, type_id: Literal["np"] = "np") -> Optional[ImageType]:
    ...


@overload
def load_image_from_file(path: Pathlike, type_id: Literal["b64"]) -> Optional[str]:
    ...


def load_image_from_file(path: Pathlike, type_id: Literal["np", "b64"] = "np") -> Optional[Union[str, ImageType]]:
    """
    Loads an image from path and passes back an encoded base64 string, a numpy array or None if file is not found
    or a conversion error occurs.

    :param path: A path to the image.
    :param type_id:  "np" or "b64".
    :return: image of desired representation
    """
    image: Optional[Union[str, ImageType]] = None
    path = path.as_posix() if isinstance(path, Path) else path

    assert is_file_extension(path, [".png", ".jpeg", ".jpg", ".tif"]), f"image type not allowed: {path}"
    assert type_id in ("np", "b64"), "type not allowed"

    try:
        if type_id == "b64":
            with open(path, "rb") as file:
                image = b64encode(file.read()).decode("utf-8")
        else:
            image = imread(path, IMREAD_COLOR)
    except (FileNotFoundError, ValueError):
        logger.info("file not found or value error: %s", path)

    return image


def load_bytes_from_pdf_file(path: Pathlike) -> bytes:
    """
    Loads a pdf file with one single page and passes back a bytes' representation of this file. Can be converted into
    a numpy or directly passed to the attr: image of Image.

    :param path: A path to a pdf file. If more pages are available, it will take the first page.
    :return: A bytes' representation of the file, width and height
    """

    assert is_file_extension(path, [".pdf"]), f"type not allowed: {path}"

    file_reader = get_pdf_file_reader(path)
    buffer = BytesIO()
    writer = get_pdf_file_writer()
    writer.addPage(file_reader.getPage(0))
    writer.write(buffer)
    return buffer.getvalue()


class LoadImageFunc(Protocol):
    """
    Protocol for typing load_image_from_file
    """

    def __call__(self, path: Pathlike) -> Optional[ImageType]:
        ...


def get_load_image_func(
    path: Pathlike,
) -> Union[LoadImageFunc, Callable[[Pathlike], bytes]]:
    """
    Return the loading function according to its file extension.

    :param path: Path to a file
    :return: The function loading the file (and converting to its desired format)
    """

    assert is_file_extension(path, [".png", ".jpeg", ".jpg", ".pdf", ".tif"]), f"image type not allowed: {path}"

    if is_file_extension(path, [".png", ".jpeg", ".jpg", ".tif"]):
        return load_image_from_file
    if is_file_extension(path, [".pdf"]):
        return load_bytes_from_pdf_file
    return NotImplemented


def maybe_path_or_pdf(path: Pathlike) -> int:
    """
    Checks if the path points to a directory or a pdf document. Returns 1 if the path points to a directory, 2
    if the path points to a pdf doc or 0, if none of the previous is true.

    :param path: A path
    :return: A value of 0,1,2
    """

    is_dir = os.path.isdir(path)
    if is_dir:
        return 1
    file_name = os.path.split(path)[1]
    is_pdf = is_file_extension(file_name, ".pdf")
    if is_pdf:
        return 2
    return 0


def load_json(path_ann: Pathlike) -> JsonDict:
    """
    Loading json file

    :param path_ann: path
    :return: dict
    """
    with open(path_ann, "r", encoding="utf-8") as file:
        json_dict = json.loads(file.read())
    return json_dict



# -*- coding: utf-8 -*-
# File: d2detect.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
D2 GeneralizedRCNN model as predictor for deepdoctection pipeline
"""

from copy import copy
from pathlib import Path
from typing import Dict, List, Literal, Mapping, Optional

from ..utils.detection_types import ImageType, Requirement
from ..utils.file_utils import (
    detectron2_available,
    get_detectron2_requirement,
    get_pytorch_requirement,
    pytorch_available,
)
from ..utils.settings import ObjectTypes, TypeOrStr, get_type
from ..utils.transform import InferenceResize
from .base import DetectionResult, ObjectDetector, PredictorBase
from .pt.ptutils import set_torch_auto_device

if pytorch_available():
    import torch
    import torch.cuda
    from torch import nn  # pylint: disable=W0611

if detectron2_available():
    from detectron2.checkpoint import DetectionCheckpointer
    from detectron2.config import CfgNode, get_cfg  # pylint: disable=W0611
    from detectron2.layers import batched_nms
    from detectron2.modeling import GeneralizedRCNN, build_model  # pylint: disable=W0611
    from detectron2.structures import Instances  # pylint: disable=W0611


def _d2_post_processing(
    predictions: Dict[str, "Instances"], nms_thresh_class_agnostic: float
) -> Dict[str, "Instances"]:
    """
    D2 postprocessing steps, so that detection outputs are aligned with outputs of other packages (e.g. Tensorpack).
    Apply a class agnostic NMS.

    :param predictions: Prediction outputs from the model.
    :param nms_thresh_class_agnostic: Nms being performed over all class predictions
    :return: filtered predictions outputs
    """
    instances = predictions["instances"]
    class_masks = torch.ones(instances.pred_classes.shape, dtype=torch.uint8)
    keep = batched_nms(instances.pred_boxes.tensor, instances.scores, class_masks, nms_thresh_class_agnostic)
    fg_instances_keep = instances[keep]
    return {"instances": fg_instances_keep}


def d2_predict_image(
    np_img: ImageType,
    predictor: "nn.Module",
    preproc_short_edge_size: int,
    preproc_max_size: int,
    nms_thresh_class_agnostic: float,
) -> List[DetectionResult]:
    """
    Run detection on one image, using the D2 model callable. It will also handle the preprocessing internally which
    is using a custom resizing within some bounds.

    :param np_img: ndarray
    :param predictor: torch nn module implemented in Detectron2
    :param preproc_short_edge_size: the short edge to resize to
    :param preproc_max_size: upper bound of one edge when resizing
    :param nms_thresh_class_agnostic: class agnostic nms threshold
    :return: list of DetectionResult
    """
    height, width = np_img.shape[:2]
    resizer = InferenceResize(preproc_short_edge_size, preproc_max_size)
    resized_img = resizer.get_transform(np_img).apply_image(np_img)
    image = torch.as_tensor(resized_img.astype("float32").transpose(2, 0, 1))

    with torch.no_grad():
        inputs = {"image": image, "height": height, "width": width}
        predictions = predictor([inputs])[0]
        predictions = _d2_post_processing(predictions, nms_thresh_class_agnostic)
    instances = predictions["instances"]
    results = [
        DetectionResult(
            box=instances[k].pred_boxes.tensor.tolist()[0],
            score=instances[k].scores.tolist()[0],
            class_id=instances[k].pred_classes.tolist()[0],
        )
        for k in range(len(instances))
    ]
    return results


class D2FrcnnDetector(ObjectDetector):
    """
    D2 Faster-RCNN implementation with all the available backbones, normalizations throughout the model
    as well as FPN, optional Cascade-RCNN and many more.

    Currently, masks are not included in the data model.

    There are no adjustment to the original implementation of Detectron2. Only one post-processing step is followed by
    the standard D2 output that takes into account of the situation that detected objects are disjoint. For more infos
    on this topic, see <https://github.com/facebookresearch/detectron2/issues/978> .

        config_path = ModelCatalog.get_full_path_configs("dd/d2/item/CASCADE_RCNN_R_50_FPN_GN.yaml")
        weights_path = ModelDownloadManager.maybe_download_weights_and_configs("item/d2_model-800000-layout.pkl")
        categories = ModelCatalog.get_profile("item/d2_model-800000-layout.pkl").categories

        d2_predictor = D2FrcnnDetector(config_path,weights_path,categories,device="cpu")

        detection_results = d2_predictor.predict(bgr_image_np_array)
    """

    def __init__(
        self,
        path_yaml: str,
        path_weights: str,
        categories: Mapping[str, TypeOrStr],
        config_overwrite: Optional[List[str]] = None,
        device: Optional[Literal["cpu", "cuda"]] = None,
    ):
        """
        Set up the predictor.

        The configuration of the model uses the full stack of build model tools of D2. For more information
        please check <https://detectron2.readthedocs.io/en/latest/tutorials/models.html#build-models-from-yacs-config>.

        :param path_yaml: The path to the yaml config. If the model is built using several config files, always use
                          the highest level .yaml file.
        :param path_weights: The path to the model checkpoint.
        :param categories: A dict with key (indices) and values (category names). Index 0 must be reserved for a
                           dummy 'BG' category. Note, that this convention is different from the builtin D2 framework,
                           where models in the model zoo are trained with 'BG' class having the highest index.
        :param config_overwrite:  Overwrite some hyperparameters defined by the yaml file with some new values. E.g.
                                 ["OUTPUT.FRCNN_NMS_THRESH=0.3","OUTPUT.RESULT_SCORE_THRESH=0.6"].
        :param device: "cpu" or "cuda". If not specified will auto select depending on what is available
        """

        self.name = "_".join(Path(path_weights).parts[-3:])
        self._categories_d2 = self._map_to_d2_categories(copy(categories))
        if config_overwrite is None:
            config_overwrite = []
        self.path_weights = path_weights
        d2_conf_list = ["MODEL.WEIGHTS", path_weights]
        for conf in config_overwrite:
            key, val = conf.split("=", maxsplit=1)
            d2_conf_list.extend([key, val])

        self.path_yaml = path_yaml
        self.categories = copy(categories)  # type: ignore
        self.config_overwrite = config_overwrite
        if device is not None:
            self.device = device
        else:
            self.device = set_torch_auto_device()
        self.cfg = self._set_config(path_yaml, d2_conf_list, device)
        self.d2_predictor = D2FrcnnDetector.set_model(self.cfg)
        self._instantiate_d2_predictor()

    @staticmethod
    def _set_config(
        path_yaml: str, d2_conf_list: List[str], device: Optional[Literal["cpu", "cuda"]] = None
    ) -> "CfgNode":
        cfg = get_cfg()
        # additional attribute with default value, so that the true value can be loaded from the configs
        cfg.NMS_THRESH_CLASS_AGNOSTIC = 0.1
        cfg.merge_from_file(path_yaml)
        cfg.merge_from_list(d2_conf_list)
        if not torch.cuda.is_available() or device == "cpu":
            cfg.MODEL.DEVICE = "cpu"
        cfg.freeze()
        return cfg

    @staticmethod
    def set_model(config: "CfgNode") -> "GeneralizedRCNN":
        """
        Build the D2 model. It uses the available builtin tools of D2

        :param config: Model config
        :return: The GeneralizedRCNN model
        """
        return build_model(config.clone()).eval()

    def _instantiate_d2_predictor(self) -> None:
        checkpointer = DetectionCheckpointer(self.d2_predictor)
        checkpointer.load(self.cfg.MODEL.WEIGHTS)

    def predict(self, np_img: ImageType) -> List[DetectionResult]:
        """
        Prediction per image.

        :param np_img: image as numpy array
        :return: A list of DetectionResult
        """
        detection_results = d2_predict_image(
            np_img,
            self.d2_predictor,
            self.cfg.INPUT.MIN_SIZE_TEST,
            self.cfg.INPUT.MAX_SIZE_TEST,
            self.cfg.NMS_THRESH_CLASS_AGNOSTIC,
        )
        return self._map_category_names(detection_results)

    def _map_category_names(self, detection_results: List[DetectionResult]) -> List[DetectionResult]:
        """
        Populating category names to detection results

        :param detection_results: list of detection results
        :return: List of detection results with attribute class_name populated
        """
        for result in detection_results:
            result.class_name = self._categories_d2[str(result.class_id)]
            if isinstance(result.class_id, int):
                result.class_id += 1
        return detection_results

    @classmethod
    def get_requirements(cls) -> List[Requirement]:
        return [get_pytorch_requirement(), get_detectron2_requirement()]

    @classmethod
    def _map_to_d2_categories(cls, categories: Mapping[str, TypeOrStr]) -> Dict[str, ObjectTypes]:
        return {str(int(k) - 1): get_type(v) for k, v in categories.items()}

    def clone(self) -> PredictorBase:
        return self.__class__(self.path_yaml, self.path_weights, self.categories, self.config_overwrite, self.device)

    def possible_categories(self) -> List[ObjectTypes]:
        return list(self.categories.values())



# -*- coding: utf-8 -*-
# File: test_box.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing the module datapoint.box
"""

from typing import List, Optional, Sequence

from numpy import asarray
from numpy.testing import assert_almost_equal, assert_array_equal
from pytest import mark, raises

from deepdoctection.datapoint import (
    BoundingBox,
    BoundingBoxError,
    crop_box_from_image,
    global_to_local_coords,
    intersection_box,
    intersection_boxes,
    local_to_global_coords,
    merge_boxes,
    rescale_coords,
)
from deepdoctection.utils.detection_types import ImageType

from .conftest import Box


class TestBoundingBox:
    """
    Testing BoundingBox methods
    """

    @staticmethod
    @mark.basic
    def test_incomplete_data_for_bounding_box(box: Box) -> None:
        """
        Testing assertion errors when box constructor receives incomplete
        data.
        :param box: Box dataclass from fixtures
        """

        # Act and Assert
        with raises(BoundingBoxError):
            BoundingBox(ulx=box.ulx, uly=box.uly, lrx=box.lrx, absolute_coords=box.absolute_coords)

        with raises(BoundingBoxError):
            BoundingBox(ulx=box.ulx, uly=box.uly, height=box.h, width=0.0, absolute_coords=box.absolute_coords)

    @staticmethod
    @mark.basic
    def test_center_area_to_list(box: Box) -> None:
        """
        Testing internal box coordinates and get_export methods
        :param box: Box dataclass from fixtures
        """

        # Arrange
        bounding_box = BoundingBox(
            ulx=box.ulx, uly=box.uly, lrx=box.lrx, lry=box.lry, absolute_coords=box.absolute_coords
        )

        # Assert
        assert bounding_box.area == box.area
        assert bounding_box.cx == box.cx
        assert bounding_box.cy == box.cy
        assert bounding_box.to_list(mode="xyxy") == [box.ulx, box.uly, box.lrx, box.lry]
        assert bounding_box.to_list(mode="xywh") == [box.ulx, box.uly, box.w, box.h]

    @staticmethod
    @mark.basic
    def test_transform(box: Box) -> None:
        """
        Testing relative <-> absolute coordinate transformation
        :param box: Box dataclass from fixtures
        """

        # Arrange
        bounding_box_absolute = BoundingBox(
            ulx=box.ulx, uly=box.uly, lrx=box.lrx, lry=box.lry, absolute_coords=box.absolute_coords
        )

        # Act
        box_relative_list = bounding_box_absolute.transform(
            image_width=box.image_width,
            image_height=box.image_height,
            absolute_coords=False,
        ).to_list(mode="xyxy")
        box_copy = bounding_box_absolute.transform(
            image_width=box.image_width, image_height=box.image_height, absolute_coords=False
        )

        # Assert
        assert box_relative_list == [box.ulx_relative, box.uly_relative, box.lrx_relative, box.lry_relative]
        assert isinstance(box_copy, BoundingBox)
        assert box_copy.absolute_coords is False


@mark.basic
@mark.parametrize(
    "box_1,box_2,expected_box",
    [
        (
            BoundingBox(absolute_coords=True, ulx=1, uly=1.5, lrx=3.5, lry=3),
            BoundingBox(absolute_coords=True, ulx=2, uly=2.5, lrx=4.5, lry=4.0),
            BoundingBox(absolute_coords=True, ulx=2, uly=2, lrx=4, lry=3),
        ),
        (
            BoundingBox(absolute_coords=True, ulx=0, uly=0, lrx=100, lry=100),
            BoundingBox(absolute_coords=True, ulx=10, uly=15, lrx=35.5, lry=30),
            BoundingBox(absolute_coords=True, ulx=10, uly=15, lrx=36, lry=30),
        ),
        (
            BoundingBox(absolute_coords=False, ulx=0.25, uly=0.4, lrx=0.6, lry=0.6),
            BoundingBox(absolute_coords=False, ulx=0.3, uly=0.55, lrx=0.4, lry=0.8),
            BoundingBox(absolute_coords=False, ulx=0.3, uly=0.55, lrx=0.4, lry=0.6),
        ),
    ],
)
def test_intersection_box(box_1: BoundingBox, box_2: BoundingBox, expected_box: BoundingBox) -> None:
    """
    Testing intersection box
    """

    # Act
    output_box = intersection_box(box_1, box_2)

    # Assert
    assert output_box == expected_box


@mark.basic
@mark.parametrize(
    "box_1,box_2,width,height,expected_box",
    [
        (
            BoundingBox(absolute_coords=True, ulx=1, uly=5.0, lrx=3.0, lry=13.0),
            BoundingBox(absolute_coords=False, ulx=0.2, uly=0.5, lrx=0.5, lry=0.7),
            10,
            20,
            BoundingBox(absolute_coords=False, ulx=0.2, uly=0.5, lrx=0.3, lry=0.65),
        ),
        (
            BoundingBox(absolute_coords=False, ulx=0, uly=0, lrx=1.0, lry=1.0),
            BoundingBox(absolute_coords=True, ulx=10, uly=15, lrx=35.5, lry=30),
            100,
            200,
            BoundingBox(absolute_coords=True, ulx=10, uly=15, lrx=36, lry=30),
        ),
    ],
)
def test_intersection_box_with_diff_abs_coords(
    box_1: BoundingBox, box_2: BoundingBox, width: float, height: float, expected_box: BoundingBox
) -> None:
    """
    Testing intersection box with different absolute coords format
    """

    # Act
    output_box = intersection_box(box_1, box_2, width, height)

    # Assert
    assert output_box == expected_box


def get_np_array_for_cropping() -> ImageType:
    """
    numpy array for cropping
    """
    return asarray([[[0, 1, 2], [3, 4, 5], [6, 7, 8]], [[9, 10, 11], [12, 13, 14], [15, 16, 17]]])


@mark.basic
@mark.parametrize(
    "np_image,crop_box,width,height,expected_np_array",
    [
        (
            get_np_array_for_cropping(),
            BoundingBox(absolute_coords=True, ulx=1, uly=1, lrx=3, lry=3),
            None,
            None,
            asarray([[[12, 13, 14], [15, 16, 17]]]),
        ),
        (
            get_np_array_for_cropping(),
            BoundingBox(absolute_coords=True, ulx=0.5, uly=1.0, lrx=1.5, lry=2.3),
            None,
            None,
            asarray([[[9, 10, 11], [12, 13, 14]]]),
        ),
        (
            get_np_array_for_cropping(),
            BoundingBox(absolute_coords=True, ulx=0, uly=0, lrx=1, lry=1),
            None,
            None,
            asarray([[[0, 1, 2]]]),
        ),
        (
            get_np_array_for_cropping(),
            BoundingBox(absolute_coords=False, ulx=0, uly=0, lrx=0.5, lry=0.5),
            2,
            3,
            asarray([[[0, 1, 2]], [[9, 10, 11]]]),
        ),
    ],
)
def test_crop_image(
    np_image: ImageType,
    crop_box: BoundingBox,
    width: Optional[float],
    height: Optional[float],
    expected_np_array: ImageType,
) -> None:
    """
    Testing func: crop_image returns np_image coorectly
    """

    # Act
    cropped_image = crop_box_from_image(np_image, crop_box, width, height)

    # Assert
    assert_array_equal(cropped_image, expected_np_array)


@mark.basic
@mark.parametrize(
    "local_box,embedding_box,expected_embedded_box",
    [
        (
            BoundingBox(absolute_coords=True, ulx=10, uly=15.5, lrx=20.0, lry=22.5),
            BoundingBox(absolute_coords=True, ulx=100, uly=150, lrx=200.0, lry=225),
            BoundingBox(absolute_coords=True, ulx=110, uly=165.5, lrx=120.0, lry=172.5),
        )
    ],
)
def test_local_to_global_coords(
    local_box: BoundingBox, embedding_box: BoundingBox, expected_embedded_box: BoundingBox
) -> None:
    """
    Testing func:  local_to_global_coords returns BoundingBox with global coords correctly
    """

    # Act
    embedded_box = local_to_global_coords(local_box, embedding_box)

    # Assert
    assert embedded_box == expected_embedded_box


@mark.basic
@mark.parametrize(
    "global_box,embedding_box,expected_local_box",
    [
        (
            BoundingBox(absolute_coords=True, ulx=100, uly=150, lrx=200, lry=200),
            BoundingBox(absolute_coords=True, ulx=50, uly=50, lrx=250, lry=225),
            BoundingBox(absolute_coords=True, ulx=50, uly=100, lrx=150, lry=150),
        ),
        (
            BoundingBox(absolute_coords=True, ulx=10, uly=20, lrx=20, lry=30),
            BoundingBox(absolute_coords=True, ulx=10, uly=25, lrx=200, lry=200),
            BoundingBox(absolute_coords=True, ulx=0, uly=0, lrx=10, lry=5),
        ),
    ],
)
def test_global_to_local_coords(
    global_box: BoundingBox, embedding_box: BoundingBox, expected_local_box: BoundingBox
) -> None:
    """
    Testing func: global_to_local_coords returns BoundingBox with local coords correclty
    """

    # Act
    local_box = global_to_local_coords(global_box, embedding_box)

    # Assert
    assert local_box == expected_local_box


@mark.basic
@mark.parametrize(
    "box_list,expected_box",
    [
        (
            [
                BoundingBox(absolute_coords=True, ulx=100, uly=150, lrx=200, lry=200),
                BoundingBox(absolute_coords=True, ulx=50, uly=50, lrx=250, lry=225),
                BoundingBox(absolute_coords=True, ulx=50, uly=100, lrx=150, lry=150),
            ],
            BoundingBox(absolute_coords=True, ulx=50, uly=50, lrx=250, lry=225),
        ),
        (
            [
                BoundingBox(absolute_coords=False, ulx=0.2, uly=0.3, lrx=0.4, lry=0.4),
                BoundingBox(absolute_coords=False, ulx=0.6, uly=0.6, lrx=0.7, lry=0.8),
            ],
            BoundingBox(absolute_coords=False, ulx=0.2, uly=0.3, lrx=0.7, lry=0.8),
        ),
    ],
)
def test_merge_boxes(box_list: List[BoundingBox], expected_box: BoundingBox) -> None:
    """
    Test func: merge_boxes returns smallest BoundingBox containing all input boxes
    """

    # Act
    merged_box = merge_boxes(*box_list)

    # Assert
    assert merged_box == expected_box


@mark.basic
@mark.parametrize(
    "box,current_total_width,current_total_height,scaled_total_width,scaled_total_height,expected_box",
    [
        (
            BoundingBox(absolute_coords=True, ulx=100, uly=150, lrx=200, lry=200),
            500,
            700,
            1000,
            2100,
            BoundingBox(absolute_coords=True, ulx=200, uly=450, lrx=400, lry=600),
        ),
        (
            BoundingBox(absolute_coords=False, ulx=0.7, uly=0.3, lrx=0.85, lry=0.4),
            500,
            700,
            1000,
            2100,
            BoundingBox(absolute_coords=False, ulx=0.7, uly=0.3, lrx=0.85, lry=0.4),
        ),
        (
            BoundingBox(absolute_coords=True, ulx=35.5, uly=47, lrx=92, lry=81),
            500,
            700,
            100,
            210,
            BoundingBox(absolute_coords=True, ulx=7.1, uly=14.1, lrx=18.4, lry=24.3),
        ),
    ],
)
def test_rescale_coords(
    box: BoundingBox,
    current_total_width: float,
    current_total_height: float,
    scaled_total_width: float,
    scaled_total_height: float,
    expected_box: BoundingBox,
) -> None:
    """
    Test func: rescale_coords returns rescaled BoundingBox, if coordinates are in absolute coords terms
    """

    # Act
    rescaled_box = rescale_coords(
        box, current_total_width, current_total_height, scaled_total_width, scaled_total_height
    )

    # Assert
    np_box = rescaled_box.to_np_array(mode="xyxy")
    np_expected_box = expected_box.to_np_array(mode="xyxy")
    assert_almost_equal(np_box, np_expected_box, decimal=3)


@mark.basic
@mark.parametrize(
    "box_1,box_2,expected_boxes",
    [
        (
            [
                BoundingBox(absolute_coords=True, ulx=10, uly=10, lrx=20, lry=70),
            ],
            [
                BoundingBox(absolute_coords=True, ulx=10, uly=10, lrx=30, lry=30),
                BoundingBox(absolute_coords=True, ulx=10, uly=40, lrx=30, lry=60),
            ],
            [
                BoundingBox(absolute_coords=True, ulx=10, uly=10, lrx=20, lry=30),
                BoundingBox(absolute_coords=True, ulx=10, uly=40, lrx=20, lry=60),
            ],
        ),
        (
            [BoundingBox(absolute_coords=False, ulx=0.1, uly=0.2, lrx=0.3, lry=0.4)],
            [BoundingBox(absolute_coords=False, ulx=0.5, uly=0.6, lrx=0.7, lry=0.8)],
            [],
        ),
    ],
)
def test_intersection_boxes(
    box_1: Sequence[BoundingBox], box_2: Sequence[BoundingBox], expected_boxes: Sequence[BoundingBox]
) -> None:
    """
    Testing intersection_boxes
    """

    # Act
    output_boxes = intersection_boxes(box_1, box_2)

    # Assert
    assert output_boxes == expected_boxes



# -*- coding: utf-8 -*-
# File: model_cascade.py

# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")

"""
This file is modified from
<https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/modeling/model_cascade.py>
"""

# pylint: disable=import-error
import tensorflow as tf
from tensorpack import tfv1
from tensorpack.tfutils import get_current_tower_context

from ..utils.box_ops import area as tf_area
from ..utils.box_ops import pairwise_iou
from .model_box import clip_boxes
from .model_frcnn import BoxProposals, FastRCNNHead, fastrcnn_outputs

# pylint: enable=import-error


class CascadeRCNNHead:
    """
    Cascade RCNN Head
    """

    def __init__(
        self, proposals, roi_func, fastrcnn_head_func, gt_targets, image_shape2d, num_categories, cfg
    ):  # pylint: disable =W0613
        """
        :param proposals: BoxProposals
        :param roi_func: a function to crop features with rois
        :param fastrcnn_head_func: the fastrcnn head to apply on the cropped features
        :param gt_targets: gt_targets
        :param image_shape2d: image shape (height,width)
        :param num_categories: number of categories
        :param cfg: config
        """

        self.cfg = cfg
        for key, value in locals().items():
            if key != "self":
                setattr(self, key, value)
        self.gt_boxes, self.gt_labels = gt_targets
        del self.gt_targets  # pylint: disable =E1101

        self.num_cascade_stages = len(cfg.CASCADE.IOUS)

        self.training = get_current_tower_context().is_training
        if self.training:

            @tf.custom_gradient
            def scale_gradient(x):
                return x, lambda dy: dy * (1.0 / self.num_cascade_stages)

            self.scale_gradient = scale_gradient
        else:
            self.scale_gradient = tf.identity

        ious = cfg.CASCADE.IOUS
        # It's unclear how to do >3 stages, so it does not make sense to implement them
        assert self.num_cascade_stages == 3, "Only 3-stage cascade was implemented!"
        with tfv1.variable_scope("cascade_rcnn_stage1"):
            H1, B1 = self.run_head(self.proposals, 0)  # pylint: disable =E1101

        with tfv1.variable_scope("cascade_rcnn_stage2"):
            B1_proposal = self.match_box_with_gt(B1, ious[1])
            H2, B2 = self.run_head(B1_proposal, 1)

        with tfv1.variable_scope("cascade_rcnn_stage3"):
            B2_proposal = self.match_box_with_gt(B2, ious[2])
            H3, B3 = self.run_head(B2_proposal, 2)
        self._cascade_boxes = [B1, B2, B3]
        self._heads = [H1, H2, H3]

    def run_head(self, proposals, stage):
        """
        Run the Fast-RCNN Head

        :param proposals: BoxProposals
        :param stage: 0, 1, 2
        :return: FastRCNNHead
                 Nx4, updated boxes
        """

        reg_weights = tf.constant(self.cfg.CASCADE.BBOX_REG_WEIGHTS[stage], dtype=tf.float32)
        pooled_feature = self.roi_func(proposals.boxes)  # N,C,S,S  # pylint: disable =E1101
        pooled_feature = self.scale_gradient(pooled_feature)
        head_feature = self.fastrcnn_head_func("head", pooled_feature, cfg=self.cfg)  # pylint: disable =E1101
        label_logits, box_logits = fastrcnn_outputs(  # pylint: disable =E1124
            "outputs", head_feature, self.num_categories, class_agnostic_regression=True  # pylint: disable =E1101
        )
        head = FastRCNNHead(proposals, box_logits, label_logits, self.gt_boxes, reg_weights, self.cfg)

        refined_boxes = head.decoded_output_boxes_class_agnostic()
        refined_boxes = clip_boxes(refined_boxes, self.image_shape2d)  # pylint: disable =E1101
        if self.training:
            refined_boxes = tf.boolean_mask(refined_boxes, tf_area(refined_boxes) > 0)
        return head, tf.stop_gradient(refined_boxes, name="output_boxes")

    def match_box_with_gt(self, boxes, iou_threshold):
        """
        Match box with ground truth

        :param boxes: Nx4
        :param iou_threshold: float
        :return: BoxProposals
        """
        if self.training:
            with tf.name_scope(f"match_box_with_gt_{iou_threshold}"):
                iou = pairwise_iou(boxes, self.gt_boxes)  # NxM
                max_iou_per_box = tf.reduce_max(iou, axis=1)  # N
                best_iou_ind = tf.argmax(iou, axis=1)  # N
                labels_per_box = tf.gather(self.gt_labels, best_iou_ind)
                fg_mask = max_iou_per_box >= iou_threshold
                fg_inds_wrt_gt = tf.boolean_mask(best_iou_ind, fg_mask)
                labels_per_box = tf.stop_gradient(labels_per_box * tf.cast(fg_mask, tf.int64))
                return BoxProposals(boxes, labels_per_box, fg_inds_wrt_gt)
        else:
            return BoxProposals(boxes)

    def losses(self):
        """
        losses
        """
        ret = []
        for idx, head in enumerate(self._heads):
            with tf.name_scope(f"cascade_loss_stage{idx + 1}"):
                ret.extend(head.losses())
        return ret

    def decoded_output_boxes(self):
        """
        decoded output boxes

        :returns:  Nx#classx4
        """
        ret = self._cascade_boxes[-1]
        ret = tf.expand_dims(ret, 1)  # class-agnostic
        return tf.tile(ret, [1, self.num_categories + 1, 1])  # pylint: disable =E1101

    def output_scores(self, name=None):
        """
        :returns: Nx#class
        """
        scores = [head.output_scores(f"cascade_scores_stage{idx + 1}") for idx, head in enumerate(self._heads)]
        return tf.multiply(tf.add_n(scores), (1.0 / self.num_cascade_stages), name=name)



# -*- coding: utf-8 -*-
# File: dataflow_builder.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for DataFlowBaseBuilder class.
"""

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Mapping, Optional, Sequence, Union

from ..dataflow import DataFlow
from ..utils.detection_types import Pathlike
from ..utils.systools import get_dataset_dir_path
from .info import DatasetCategories


class DataFlowBaseBuilder(ABC):
    """
    Abstract base class for building the dataflow of a dataset.

    DataFlowBase has an abstract `build` that returns the dataflow of a dataset. The dataflow should be
    designed in such a way that each data point is already mapped in the form of the core data model and thus
    corresponds to a `datapoint.Image` instance. Any characteristics can be passed as arguments and implemented,
    which influence the return of the dataflow. These include, for example, the `split`, `max_datapoints` but also
    specific further transformations, such as cutting and returning an annotation as a sub image. Within this method,
    checks and consistency checks should also be carried out so that a curated data flow is available as return value.
    Such specific transformations should be implemented by transferring a value of the argument `build_mode`.
    """

    def __init__(
        self,
        location: Pathlike,
        annotation_files: Optional[Mapping[str, Union[str, Sequence[str]]]] = None,
    ):
        """
        :param location: Relative path of the physical dataset.
        :param annotation_files: Dict of annotation files e.g. depending on the split.
        """
        self.location = location
        if annotation_files is None:
            annotation_files = {}
        self.annotation_files = annotation_files
        self._categories: Optional[DatasetCategories] = None
        self._splits: Mapping[str, str] = {}

    @property
    def categories(self) -> DatasetCategories:
        """
        categories
        """
        if self._categories is not None:
            return self._categories
        raise ValueError(f"DatasetCategories not defined for {self.__class__.__name__}")

    @categories.setter
    def categories(self, categories: DatasetCategories) -> None:
        """
        categories setter
        """
        self._categories = categories

    def get_split(self, key: str) -> str:
        """
        split value
        """
        return self._splits[key]

    @property
    def splits(self) -> Mapping[str, str]:
        """
        splits
        """
        return self._splits

    @splits.setter
    def splits(self, splits: Mapping[str, str]) -> None:
        """
        set splits
        """
        self._splits = splits

    def get_workdir(self) -> Path:
        """
        Get the absolute path to the locally physically stored dataset.

        :return: local workdir
        """
        return get_dataset_dir_path() / self.location

    @abstractmethod
    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Consult the docstring w.r.t `DataFlowBaseBuilder`.

        :param kwargs: A custom set of arguments/values
        :return: dataflow
        """
        raise NotImplementedError

    def get_annotation_file(self, split: str) -> str:
        """Get single annotation file."""
        split_file = self.annotation_files[split]
        if isinstance(split_file, str):
            return split_file
        raise TypeError("Unsupported get method for sequence type splits")



# -*- coding: utf-8 -*-
# File: layout.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for layout pipeline component
"""
from typing import Optional

import numpy as np

from ..datapoint.image import Image
from ..extern.base import ObjectDetector, PdfMiner
from ..utils.detection_types import JsonDict
from ..utils.transform import PadTransform
from .base import PredictorPipelineComponent
from .registry import pipeline_component_registry


@pipeline_component_registry.register("ImageLayoutService")
class ImageLayoutService(PredictorPipelineComponent):
    """
    Pipeline component for determining the layout. Which layout blocks are determined depends on the Detector and thus
    usually on the data set on which the Detector was pre-trained. If the Detector has been trained on Publaynet, these
    are layouts such as text, title, table, list and figure. If the Detector has been trained on DocBank, these are
    rather Abstract, Author, Caption, Equation, Figure, Footer, List, Paragraph, Reference, Section, Table, Title.

    The component is usually at the beginning of the pipeline. Cropping of the layout blocks can be selected to simplify
    further processing.

    **Example**

            d_items = TPFrcnnDetector(item_config_path, item_weights_path, {"1": "ROW", "2": "COLUMNS"})
            item_component = ImageLayoutService(d_items)
    """

    def __init__(
        self,
        layout_detector: ObjectDetector,
        to_image: bool = False,
        crop_image: bool = False,
        padder: Optional[PadTransform] = None,
    ):
        """
        :param layout_detector: object detector
        :param to_image: Generate an image for each detected block, e.g. populate `ImageAnnotation.image`. Useful,
                         if you want to process only some blocks in a subsequent pipeline component.
        :param crop_image: Do not only populate `ImageAnnotation.image` but also crop the detected block according
                           to its bounding box and populate the resulting sub image to
                           `ImageAnnotation.image.image`.
        """
        self.to_image = to_image
        self.crop_image = crop_image
        self.padder = padder
        super().__init__(self._get_name(layout_detector.name), layout_detector)

    def serve(self, dp: Image) -> None:
        if dp.image is None:
            raise ValueError("image cannot be None")
        np_image = dp.image
        if self.padder:
            np_image = self.padder.apply_image(np_image)
        detect_result_list = self.predictor.predict(np_image)  # type: ignore
        if self.padder and detect_result_list:
            boxes = np.array([detect_result.box for detect_result in detect_result_list])
            boxes_orig = self.padder.inverse_apply_coords(boxes)
            for idx, detect_result in enumerate(detect_result_list):
                detect_result.box = boxes_orig[idx, :].tolist()

        for detect_result in detect_result_list:
            self.dp_manager.set_image_annotation(detect_result, to_image=self.to_image, crop_image=self.crop_image)

    def get_meta_annotation(self) -> JsonDict:
        assert isinstance(self.predictor, (ObjectDetector, PdfMiner))
        return dict(
            [
                ("image_annotations", self.predictor.possible_categories()),
                ("sub_categories", {}),
                ("relationships", {}),
                ("summaries", []),
            ]
        )

    @staticmethod
    def _get_name(predictor_name: str) -> str:
        return f"image_{predictor_name}"

    def clone(self) -> "PredictorPipelineComponent":
        predictor = self.predictor.clone()
        padder_clone = None
        if self.padder:
            padder_clone = self.padder.clone()
        if not isinstance(predictor, ObjectDetector):
            raise ValueError(f"predictor must be of type ObjectDetector, but is of type {type(predictor)}")
        return self.__class__(predictor, self.to_image, self.crop_image, padder_clone)
